from __future__ import annotations

from typing import Literal

from prettyqt import core, multimedia
from prettyqt.utils import bidict


ErrorStr = Literal["none", "camera"]

ERROR: bidict[ErrorStr, multimedia.QCamera.Error] = bidict(
    none=multimedia.QCamera.Error.NoError,
    camera=multimedia.QCamera.Error.CameraError,
)

ExposureModeStr = Literal[
    "auto",
    "manual",
    "portrait",
    "night",
    "sports",
    "snow",
    "beach",
    "action",
    "landscape",
    "night_portrait",
    "threatre",
    "sunset",
    "steady_photo",
    "fireworks",
    "party",
    "candle_light",
    "barcode",
]

EXPOSURE_MODE: bidict[ExposureModeStr, multimedia.QCamera.ExposureMode] = bidict(
    auto=multimedia.QCamera.ExposureMode.ExposureAuto,
    manual=multimedia.QCamera.ExposureMode.ExposureManual,
    portrait=multimedia.QCamera.ExposureMode.ExposurePortrait,
    night=multimedia.QCamera.ExposureMode.ExposureNight,
    sports=multimedia.QCamera.ExposureMode.ExposureSports,
    snow=multimedia.QCamera.ExposureMode.ExposureSnow,
    beach=multimedia.QCamera.ExposureMode.ExposureBeach,
    action=multimedia.QCamera.ExposureMode.ExposureAction,
    landscape=multimedia.QCamera.ExposureMode.ExposureLandscape,
    night_portrait=multimedia.QCamera.ExposureMode.ExposureNightPortrait,
    threatre=multimedia.QCamera.ExposureMode.ExposureTheatre,
    sunset=multimedia.QCamera.ExposureMode.ExposureSunset,
    steady_photo=multimedia.QCamera.ExposureMode.ExposureSteadyPhoto,
    fireworks=multimedia.QCamera.ExposureMode.ExposureFireworks,
    party=multimedia.QCamera.ExposureMode.ExposureParty,
    candle_light=multimedia.QCamera.ExposureMode.ExposureCandlelight,
    barcode=multimedia.QCamera.ExposureMode.ExposureBarcode,
)

FeatureStr = Literal[
    "color_temperature",
    "exposure_compensation",
    "iso_sensitivity",
    "manual_exposure_time",
    "custom_focus_point",
    "focus_distance",
]

FEATURE: bidict[FeatureStr, multimedia.QCamera.Feature] = bidict(
    none=multimedia.QCamera.Feature(0),
    color_temperature=multimedia.QCamera.Feature.ColorTemperature,
    exposure_compensation=multimedia.QCamera.Feature.ExposureCompensation,
    iso_sensitivity=multimedia.QCamera.Feature.IsoSensitivity,
    manual_exposure_time=multimedia.QCamera.Feature.ManualExposureTime,
    custom_focus_point=multimedia.QCamera.Feature.CustomFocusPoint,
    focus_distance=multimedia.QCamera.Feature.FocusDistance,
)

FlashModeStr = Literal["off", "on", "auto"]

FLASH_MODE: bidict[FlashModeStr, multimedia.QCamera.FlashMode] = bidict(
    off=multimedia.QCamera.FlashMode.FlashOff,
    on=multimedia.QCamera.FlashMode.FlashOn,
    auto=multimedia.QCamera.FlashMode.FlashAuto,
)

FocusModeStr = Literal[
    "auto",
    "auto_near",
    "auto_far",
    "focal",
    "infinity",
    "manual",
]

FOCUS_MODE: bidict[FocusModeStr, multimedia.QCamera.FocusMode] = bidict(
    auto=multimedia.QCamera.FocusMode.FocusModeAuto,
    auto_near=multimedia.QCamera.FocusMode.FocusModeAutoNear,
    auto_far=multimedia.QCamera.FocusMode.FocusModeAutoFar,
    focal=multimedia.QCamera.FocusMode.FocusModeHyperfocal,
    infinity=multimedia.QCamera.FocusMode.FocusModeInfinity,
    manual=multimedia.QCamera.FocusMode.FocusModeManual,
)

TorchModeStr = Literal["off", "on", "auto"]

TORCH_MODE: bidict[TorchModeStr, multimedia.QCamera.TorchMode] = bidict(
    off=multimedia.QCamera.TorchMode.TorchOff,
    on=multimedia.QCamera.TorchMode.TorchOn,
    auto=multimedia.QCamera.TorchMode.TorchAuto,
)

WhiteBalanceModeStr = Literal[
    "auto",
    "manual",
    "sunlight",
    "cloudy",
    "shade",
    "tungsten",
    "fluorescent",
    "flash",
    "sunset",
]

WHITE_BALANCE: bidict[WhiteBalanceModeStr, multimedia.QCamera.WhiteBalanceMode] = bidict(
    auto=multimedia.QCamera.WhiteBalanceMode.WhiteBalanceAuto,
    manual=multimedia.QCamera.WhiteBalanceMode.WhiteBalanceManual,
    sunlight=multimedia.QCamera.WhiteBalanceMode.WhiteBalanceSunlight,
    cloudy=multimedia.QCamera.WhiteBalanceMode.WhiteBalanceCloudy,
    shade=multimedia.QCamera.WhiteBalanceMode.WhiteBalanceShade,
    tungsten=multimedia.QCamera.WhiteBalanceMode.WhiteBalanceTungsten,
    fluorescent=multimedia.QCamera.WhiteBalanceMode.WhiteBalanceFluorescent,
    flash=multimedia.QCamera.WhiteBalanceMode.WhiteBalanceFlash,
    sunset=multimedia.QCamera.WhiteBalanceMode.WhiteBalanceSunset,
)


class Camera(core.ObjectMixin, multimedia.QCamera):
    def set_exposure_mode(self, mode: ExposureModeStr | multimedia.QCamera.ExposureMode):
        """Set the exposure mode.

        Args:
            mode: exposure mode
        """
        self.setExposureMode(EXPOSURE_MODE.get_enum_value(mode))

    def get_exposure_mode(self) -> ExposureModeStr:
        """Return current exposure mode.

        Returns:
            exposure mode
        """
        return EXPOSURE_MODE.inverse[self.exposureMode()]

    def set_torch_mode(self, mode: TorchModeStr | multimedia.QCamera.TorchMode):
        """Set the torch mode.

        Args:
            mode: torch mode
        """
        self.setTorchMode(TORCH_MODE.get_enum_value(mode))

    def get_torch_mode(self) -> TorchModeStr:
        """Return current torch mode.

        Returns:
            torch mode
        """
        return TORCH_MODE.inverse[self.torchMode()]

    def set_flash_mode(self, mode: FlashModeStr | multimedia.QCamera.FlashMode):
        """Set the flash mode.

        Args:
            mode: flash mode
        """
        self.setFlashMode(FLASH_MODE.get_enum_value(mode))

    def get_flash_mode(self) -> FlashModeStr:
        """Return current flash mode.

        Returns:
            flash mode
        """
        return FLASH_MODE.inverse[self.flashMode()]

    def set_white_balance_mode(
        self, mode: WhiteBalanceModeStr | multimedia.QCamera.WhiteBalanceMode
    ):
        """Set the white balance mode.

        Args:
            mode: white balance mode
        """
        self.setWhiteBalanceMode(WHITE_BALANCE.get_enum_value(mode))

    def get_white_balance_mode(self) -> WhiteBalanceModeStr:
        """Return current white balance mode.

        Returns:
            white balance mode
        """
        return WHITE_BALANCE.inverse[self.whiteBalanceMode()]

    def get_supported_features(self) -> list[FeatureStr]:
        return FEATURE.get_list(self.supportedFeatures())

    def get_error(self) -> ErrorStr:
        """Return current error state.

        Returns:
            error state
        """
        return ERROR.inverse[self.error()]

    def get_camera_format(self) -> multimedia.CameraFormat:
        return multimedia.CameraFormat(self.cameraFormat())


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    cam = Camera()
    cam.supportedFeatures()
    print(type(cam.supportedFeatures()).mro())



from __future__ import annotations

from prettyqt import constants, core
from prettyqt.qt import QtGui
from prettyqt.utils import bidict


ACTION = bidict(
    click=QtGui.QInputMethod.Action.Click,
    context_menu=QtGui.QInputMethod.Action.ContextMenu,
)


class InputMethod(core.ObjectMixin):
    def __init__(self, item: QtGui.QInputMethod):
        self.item = item

    def __getattr__(self, val):
        return getattr(self.item, val)

    def get_anchor_rectangle(self) -> core.RectF:
        return core.RectF(self.anchorRectangle())

    def get_cursor_rectangle(self) -> core.RectF:
        return core.RectF(self.cursorRectangle())

    def get_input_item_clip_rectangle(self) -> core.RectF:
        return core.RectF(self.inputItemClipRectangle())

    def get_input_item_rectangle(self) -> core.RectF:
        return core.RectF(self.inputItemRectangle())

    def get_keyboard_rectangle(self) -> core.RectF:
        return core.RectF(self.keyboardRectangle())

    def get_locale(self) -> core.Locale:
        return core.Locale(self.locale())

    def get_input_direction(self) -> constants.LayoutDirectionStr:
        return constants.LAYOUT_DIRECTION.inverse[self.inputDirection()]

    @classmethod
    def query_focus_object(cls, query: constants.InputMethodQueryStr, argument):
        return cls.queryFocusObject(constants.INPUT_METHOD_QUERY[query], argument)


if __name__ == "__main__":
    from prettyqt import gui

    app = gui.app()
    method = app.get_input_method()



# credits to Psygnal.

"""MutableSequence that emits events when altered.

Note For Developers
===================

Be cautious when re-implementing typical list-like methods here (e.g. extend,
pop, clear, etc...).  By not re-implementing those methods, we force ALL "CRUD"
(create, read, update, delete) operations to go through a few key methods
defined by the abc.MutableSequence interface, where we can emit the necessary
events.

Specifically:

- `insert` = "create" : add a new item/index to the list
- `__getitem__` = "read" : get the value of an existing index
- `__setitem__` = "update" : update the value of an existing index
- `__delitem__` = "delete" : remove an existing index from the list

All of the additional list-like methods are provided by the MutableSequence
interface, and call one of those 4 methods.  So if you override a method, you
MUST make sure that all the appropriate events are emitted.  (Tests should
cover this in test_evented_list.py)
"""
from __future__ import annotations

from collections.abc import Iterable, MutableSequence
from typing import Any, TypeVar, cast

from prettyqt import core


_T = TypeVar("_T")


class Signals(core.Object):
    inserting = core.Signal(int)  # idx
    inserted = core.Signal(int, object)  # (idx, value)
    removing = core.Signal(int)  # idx
    removed = core.Signal(int, object)  # (idx, value)
    moving = core.Signal(int, int)  # (src_idx, dest_idx)
    moved = core.Signal(int, int, object)  # (src_idx, dest_idx, value)
    changed = core.Signal(object, object, object)  # (int | slice, old, new)
    reordered = core.Signal()


class SignalList(MutableSequence[_T]):
    def __init__(
        self,
        data: Iterable[_T] = (),
        *,
        hashable: bool = True,
    ):
        super().__init__()
        self._data: list[_T] = []
        self._hashable = hashable
        self.signals = Signals()
        self.extend(data)

    def insert(self, index: int, value: _T) -> None:
        """Insert `value` before index."""
        _value = self._pre_insert(value)
        self.signals.inserting.emit(index)
        self._data.insert(index, _value)
        self.signals.inserted.emit(index, value)

    def __getitem__(self, key: int | slice) -> _T | SignalList[_T]:
        result = self._data[key]
        return self.__newlike__(result) if isinstance(result, list) else result

    def __setitem__(self, key: int | slice, value: _T | Iterable[_T]) -> None:
        old = self._data[key]
        if value is old:
            return
        if isinstance(key, slice):
            if not isinstance(value, Iterable):
                raise TypeError("Can only assign an iterable to slice")
            value = [self._pre_insert(v) for v in value]  # before we mutate the list
        else:
            value = self._pre_insert(cast("_T", value))
        self._data[key] = value
        self.signals.changed.emit(key, old, value)

    def __delitem__(self, key: int | slice) -> None:
        """Delete self[key]."""
        # delete from the end
        for parent, index in sorted(self._delitem_indices(key), reverse=True):
            parent.signals.removing.emit(index)
            item = parent._data.pop(index)
            self.signals.removed.emit(index, item)

    def _delitem_indices(self, key: int | slice) -> Iterable[tuple[SignalList[_T], int]]:
        # returning (self, int) allows subclasses to pass nested members
        match key:
            case int():
                yield (self, key if key >= 0 else key + len(self))
            case slice():
                yield from ((self, i) for i in range(*key.indices(len(self))))
            case _:
                raise TypeError(key)

    def _pre_insert(self, value: _T) -> _T:
        """Validate and or modify values prior to inserted."""
        return value

    def __newlike__(self, iterable: Iterable[_T]) -> SignalList[_T]:
        """Return new instance of same class."""
        return self.__class__(iterable)

    def copy(self) -> SignalList[_T]:
        """Return a shallow copy of the list."""
        return self.__newlike__(self)

    def __add__(self, other: Iterable[_T]) -> SignalList[_T]:
        """Add other to self, return new object."""
        copy = self.copy()
        copy.extend(other)
        return copy

    def __iadd__(self, other: Iterable[_T]) -> SignalList[_T]:
        """Add other to self in place (self += other)."""
        self.extend(other)
        return self

    def __radd__(self, other: list) -> list:
        """Reflected add (other + self).  Cast self to list."""
        return other + list(self)

    def __len__(self) -> int:
        """Return len(self)."""
        return len(self._data)

    def __repr__(self) -> str:
        """Return repr(self)."""
        return f"{type(self).__name__}({self._data})"

    def __eq__(self, other: Any) -> bool:
        """Return self==value."""
        return self._data == other

    def __hash__(self) -> int:
        """Return hash(self)."""
        # it's important to add this to allow this object to be hashable
        # given that we've also reimplemented __eq__
        if self._hashable:
            return id(self)
        name = self.__class__.__name__
        raise TypeError(f"unhashable type: {name!r}.")

    def reverse(self, *, emit_individual_events: bool = False) -> None:
        """Reverse list *IN PLACE*."""
        if emit_individual_events:
            super().reverse()
        else:
            self._data.reverse()
        self.signals.reordered.emit()

    def move(
        self, src_index: int, dest_index: int = 0, emit_reordered: bool = True
    ) -> bool:
        """Insert object at `src_index` before `dest_index`.

        Both indices refer to the list prior to any object removal (pre-move space).
        """
        if dest_index < 0:
            dest_index += len(self) + 1
        if dest_index in (src_index, src_index + 1):
            # this is a no-op
            return False

        self.signals.moving.emit(src_index, dest_index)
        item = self._data.pop(src_index)
        if dest_index > src_index:
            dest_index -= 1
        self._data.insert(dest_index, item)
        self.signals.moved.emit(src_index, dest_index, item)
        if emit_reordered:
            self.signals.reordered.emit()
        return True

    def move_multiple(self, sources: Iterable[int | slice], dest_index: int = 0) -> int:
        """Move a batch of `sources` indices, to a single destination.

        Note, if `dest_index` is higher than any of the `sources`, then
        the resulting position of the moved objects after the move operation
        is complete will be lower than `dest_index`.

        Arguments:
            sources:  A sequence of indices
            dest_index: The destination index. All sources will be inserted before this
                        index (in pre-move space), by default 0... which has the effect
                        of "bringing to front" everything in `sources`, or acting as a
                        "reorder" method if `sources` contains all indices.

        Returns:
            The number of successful move operations completed.

        Raises:
            TypeError
                If the destination index is a slice, or any of the source indices
                are not `int` or `slice`.
        """
        # calling list here makes sure that there are no index errors up front
        move_plan = list(self._move_plan(sources, dest_index))

        # don't assume index adjacency ... so move objects one at a time
        # this *could* be simplified with an intermediate list ... but this way
        # allows any views (such as QtViews) to update themselves more easily.
        # If this needs to be changed in the future for performance reasons,
        # then the associated QtListView will need to changed from using
        # `beginMoveRows` & `endMoveRows` to using `layoutAboutToBeChanged` &
        # `layoutChanged` while *manually* updating model indices with
        # `changePersistentIndexList`.  That becomes much harder to do with
        # nested tree-like models.
        for src, dest in move_plan:
            self.move(src, dest, emit_reordered=False)

        self.signals.reordered.emit()
        return len(move_plan)

    def _move_plan(
        self, sources: Iterable[int | slice], dest_index: int
    ) -> Iterable[tuple[int, int]]:
        """Yield prepared indices for a multi-move.

        Given a set of `sources` from anywhere in the list,
        and a single `dest_index`, this function computes and yields
        `(from_index, to_index)` tuples that can be used sequentially in
        single move operations.  It keeps track of what has moved where and
        updates the source and destination indices to reflect the model at each
        point in the process.

        This is useful for a drag-drop operation with a QtModel/View.

        Arguments:
            sources: An iterable of tuple[int] that should be moved to `dest_index`.
            dest_index: The destination for sources.
        """
        if isinstance(dest_index, slice):
            raise TypeError("Destination index may not be a slice")  # pragma: no cover

        to_move: list[int] = []
        for idx in sources:
            match idx:
                case slice():
                    to_move.extend(list(range(*idx.indices(len(self)))))
                case int():
                    to_move.append(idx)
                case _:
                    raise TypeError(idx)

        to_move = list(dict.fromkeys(to_move))

        if dest_index < 0:
            dest_index += len(self) + 1

        d_inc = 0
        popped: list[int] = []
        for i, src in enumerate(to_move):
            if src != dest_index:
                src -= sum(x <= src for x in popped)
                if src >= dest_index:
                    src += i
                yield src, dest_index + d_inc
            popped.append(src)
            if dest_index <= src:
                d_inc += 1


if __name__ == "__main__":
    import logging

    from prettyqt import debugging, widgets

    app = widgets.app()
    container = SignalList([0])
    with app.debug_mode():
        stalker = debugging.Stalker(container.signals, log_level=logging.DEBUG)
        stalker.hook()
        container.signals.changed.connect(print)
        container[0] = 2
        app.processEvents()



from __future__ import annotations

from typing import Literal

from prettyqt import statemachine
from prettyqt.utils import bidict


ChildModeStr = Literal["exclusive", "parallel"]

CHILD_MODE: bidict[ChildModeStr, statemachine.QState.ChildMode] = bidict(
    exclusive=statemachine.QState.ChildMode.ExclusiveStates,
    parallel=statemachine.QState.ChildMode.ParallelStates,
)

RestorePolicyStr = Literal["dont_restore", "restore"]

RESTORE_POLICY: bidict[RestorePolicyStr, statemachine.QState.RestorePolicy] = bidict(
    dont_restore=statemachine.QState.RestorePolicy.DontRestoreProperties,
    restore=statemachine.QState.RestorePolicy.RestoreProperties,
)


class StateMixin(statemachine.AbstractStateMixin):
    def set_child_mode(self, mode: ChildModeStr | statemachine.QState.ChildMode):
        """Set child mode to use.

        Args:
            mode: child mode to use
        """
        self.setChildMode(CHILD_MODE.get_enum_value(mode))

    def get_child_mode(self) -> ChildModeStr:
        """Return current child mode.

        Returns:
            child mode
        """
        return CHILD_MODE.inverse[self.childMode()]


class State(StateMixin, statemachine.QState):
    pass


if __name__ == "__main__":
    reg = State()



from __future__ import annotations

from typing import Literal

from prettyqt import constants, gui, widgets
from prettyqt.utils import datatypes


LayoutStr = Literal["grid", "horizontal", "vertical", "anchor"]


class GraphicsWidgetMixin(widgets.GraphicsObjectMixin, widgets.GraphicsLayoutItemMixin):
    @property
    def box(self):
        return self.layout()

    @box.setter
    def box(self, layout):
        self.set_layout(layout)

    def set_layout(
        self, layout: LayoutStr | widgets.QGraphicsLayout
    ) -> widgets.QGraphicsLayout:
        match layout:
            case widgets.QGraphicsLayout():
                layout = layout
            case "grid":
                layout = widgets.GraphicsGridLayout()
            case "anchor":
                layout = widgets.GraphicsAnchorLayout()
            case "horizontal" | "vertical":
                layout = widgets.GraphicsLinearLayout(layout)
            case _:
                raise ValueError(f"Invalid Layout {layout}")
        self.setLayout(layout)
        return layout

    def set_focus_policy(self, policy: constants.FocusPolicyStr | constants.FocusPolicy):
        """Set the way the widget accepts keyboard focus.

        Args:
            policy: Focus policy
        """
        self.setFocusPolicy(constants.FOCUS_POLICY.get_enum_value(policy))

    def get_focus_policy(self) -> constants.FocusPolicyStr:
        """Return way the widget accepts keyboard focus.

        Returns:
            str: Focus policy
        """
        return constants.FOCUS_POLICY.inverse[self.focusPolicy()]

    def window_frame_section_at(
        self, point: datatypes.PointType
    ) -> constants.WindowFrameSectionStr:
        """Return the window frame section at given position.

        Returns:
            str: Window frame section
        """
        section = self.windowFrameSectionAt(datatypes.to_point(point))
        return constants.WINDOW_FRAME_SECTION.inverse[section]

    def get_font(self) -> gui.Font:
        return gui.Font(self.font())

    def get_palette(self) -> gui.Palette:
        return gui.Palette(self.palette())


class GraphicsWidget(GraphicsWidgetMixin, widgets.QGraphicsWidget):
    """The base class for all widget items in a QGraphicsScene."""


if __name__ == "__main__":
    widget = GraphicsWidget()
    widget.set_layout("vertical")



from __future__ import annotations

from prettyqt import core, qthelp, widgets
from prettyqt.utils import datatypes


class HelpSearchResultWidget(widgets.WidgetMixin):  # , qthelp.QHelpSearchResultWidget):
    """Text browser to display search results."""

    def __init__(self, item: qthelp.QHelpSearchResultWidget):
        self.item = item

    def __getattr__(self, val):
        return getattr(self.item, val)

    @classmethod
    def setup_example(self):
        core_engine = qthelp.HelpEngineCore("test")
        engine = qthelp.HelpSearchEngine(core_engine)
        widget = engine.get_result_widget()
        widget.engine = engine
        widget.core_engine = core_engine
        return widget

    def get_link_at(self, point: datatypes.PointType) -> core.Url:
        if isinstance(point, tuple):
            point = core.Point(*point)
        return core.Url(self.linkAt(point))


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    core_engine = qthelp.HelpEngineCore("test")
    engine = qthelp.HelpSearchEngine(core_engine)
    widget = engine.get_result_widget()
    widget.show()
    app.exec()



from __future__ import annotations

from prettyqt import core, location


class GeoRoutingManager(core.ObjectMixin):
    """Support for geographic routing operations."""

    def __init__(self, item: location.QGeoRoutingManager):
        self.item = item

    def __getattr__(self, val):
        return getattr(self.item, val)

    def get_locale(self) -> core.Locale:
        return core.Locale(self.item.locale())

    def get_supported_feature_types(
        self,
    ) -> list[location.georouterequest.FeatureTypeStr]:
        return location.georouterequest.FEATURE_TYPES.get_list(
            self.item.supportedFeatureTypes()
        )

    def get_supported_feature_weights(
        self,
    ) -> list[location.georouterequest.FeatureWeightStr]:
        return location.georouterequest.FEATURE_WEIGHTS.get_list(
            self.item.supportedFeatureWeights()
        )

    def get_supported_maneuver_details(
        self,
    ) -> list[location.georouterequest.ManeuverDetailStr]:
        return location.georouterequest.MANEUVER_DETAIL.get_list(
            self.item.supportedManeuverDetails()
        )

    def get_supported_route_optimizations(
        self,
    ) -> list[location.georouterequest.RouteOptimizationStr]:
        return location.georouterequest.ROUTE_OPTIMIZATION.get_list(
            self.item.supportedRouteOptimizations()
        )

    def get_supported_segment_details(
        self,
    ) -> list[location.georouterequest.SegmentDetailStr]:
        return location.georouterequest.SEGMENT_DETAIL.get_list(
            self.item.supportedSegmentDetails()
        )

    def get_supported_travel_modes(self) -> list[location.georouterequest.TravelModeStr]:
        return location.georouterequest.TRAVEL_MODE.get_list(
            self.item.supportedTravelModes()
        )



from __future__ import annotations

from prettyqt import core
from prettyqt.qt import QtWidgets


class GraphicsEffectMixin(core.ObjectMixin):
    pass


class GraphicsEffect(GraphicsEffectMixin, QtWidgets.QGraphicsEffect):
    """The base class for all graphics effects."""



from __future__ import annotations

from prettyqt import core, gui
from prettyqt.qt import QtSvg
from prettyqt.utils import datatypes


class SvgGenerator(gui.PaintDeviceMixin, QtSvg.QSvgGenerator):
    def get_viewbox(self) -> core.Rect:
        return core.Rect(self.viewBox())

    def get_viewboxf(self) -> core.RectF:
        return core.RectF(self.viewBoxF())

    def get_size(self) -> core.Size:
        return core.Size(self.size())

    def set_size(self, size: datatypes.SizeType | core.QSizeF):
        self.setSize(datatypes.to_size(size))



from __future__ import annotations

from prettyqt import core, qthelp


class HelpSearchEngine(core.ObjectMixin, qthelp.QHelpSearchEngine):
    """Access to widgets reusable to integrate fulltext search."""

    def search_results(self, start: int, end: int) -> list[qthelp.HelpSearchResult]:
        return [qthelp.HelpSearchResult(i) for i in self.searchResults(start, end)]

    def get_result_widget(self) -> qthelp.HelpSearchResultWidget:
        return qthelp.HelpSearchResultWidget(self.resultWidget())


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    core_engine = qthelp.HelpEngineCore("test")
    engine = HelpSearchEngine(core_engine)
    widget = engine.get_result_widget()
    widget.show()
    app.exec()



from __future__ import annotations

import datetime

from typing import Literal

from prettyqt import core
from prettyqt.utils import bidict, datatypes, get_repr


FileErrorStr = Literal[
    "none",
    "read",
    "write",
    "fatal",
    "resource",
    "open",
    "abort",
    "time_out",
    "unspecified",
    "remove",
    "rename",
    "position",
    "resize",
    "permissions",
    "copy",
]

FILE_ERROR: bidict[FileErrorStr, core.QFileDevice.FileError] = bidict(
    none=core.QFileDevice.FileError.NoError,
    read=core.QFileDevice.FileError.ReadError,
    write=core.QFileDevice.FileError.WriteError,
    fatal=core.QFileDevice.FileError.FatalError,
    resource=core.QFileDevice.FileError.ResourceError,
    open=core.QFileDevice.FileError.OpenError,
    abort=core.QFileDevice.FileError.AbortError,
    time_out=core.QFileDevice.FileError.TimeOutError,
    unspecified=core.QFileDevice.FileError.UnspecifiedError,
    remove=core.QFileDevice.FileError.RemoveError,
    rename=core.QFileDevice.FileError.RenameError,
    position=core.QFileDevice.FileError.PositionError,
    resize=core.QFileDevice.FileError.ResizeError,
    permissions=core.QFileDevice.FileError.PermissionsError,
    copy=core.QFileDevice.FileError.CopyError,
)

FileTimeStr = Literal["access", "birth", "metadata_change", "modification"]

FILE_TIME: bidict[FileTimeStr, core.QFileDevice.FileTime] = bidict(
    access=core.QFileDevice.FileTime.FileAccessTime,
    birth=core.QFileDevice.FileTime.FileBirthTime,
    metadata_change=core.QFileDevice.FileTime.FileMetadataChangeTime,
    modification=core.QFileDevice.FileTime.FileModificationTime,
)

PermissionStr = Literal[
    "read_owner",
    "write_owner",
    "exe_owner",
    "read_user",
    "write_user",
    "exe_user",
    "read_group",
    "write_group",
    "exe_group",
    "read_other",
    "write_other",
    "exe_other",
]

PERMISSIONS: bidict[PermissionStr, core.QFileDevice.Permission] = bidict(
    read_owner=core.QFileDevice.Permission.ReadOwner,
    write_owner=core.QFileDevice.Permission.WriteOwner,
    exe_owner=core.QFileDevice.Permission.ExeOwner,
    read_user=core.QFileDevice.Permission.ReadUser,
    write_user=core.QFileDevice.Permission.WriteUser,
    exe_user=core.QFileDevice.Permission.ExeUser,
    read_group=core.QFileDevice.Permission.ReadGroup,
    write_group=core.QFileDevice.Permission.WriteGroup,
    exe_group=core.QFileDevice.Permission.ExeGroup,
    read_other=core.QFileDevice.Permission.ReadOther,
    write_other=core.QFileDevice.Permission.WriteOther,
    exe_other=core.QFileDevice.Permission.ExeOther,
)


class FileDeviceMixin(core.IODeviceMixin):
    """Interface for reading from and writing to open files."""

    def __repr__(self):
        return get_repr(self, self.fileName())

    def __str__(self):
        return self.fileName()

    # def __fspath__(self) -> str:
    #     return self.fileName()

    def get_permissions(self) -> list[PermissionStr]:
        return PERMISSIONS.get_list(self.permissions())

    def set_file_time(
        self,
        file_time: datatypes.DateTimeType,
        typ: FileTimeStr | core.QFileDevice.FileTime,
    ) -> bool:
        """Set file time.

        Args:
            file_time: file time to set
            typ: file time type
        """
        file_time = datatypes.to_datetime(file_time)
        return self.setFileTime(file_time, FILE_TIME.get_enum_value(typ))  # type: ignore

    def get_file_time(
        self, typ: FileTimeStr | core.QFileDevice.FileTime
    ) -> datetime.datetime | None:
        """Return current file time.

        Returns:
            file time
        """
        if date := self.fileTime(FILE_TIME.get_enum_value(typ)):
            return date.toPython()  # type: ignore
        return None

    def get_error(self) -> FileErrorStr:
        """Return file error status.

        Returns:
            file error status
        """
        return FILE_ERROR.inverse[self.error()]

    def is_readable(self) -> bool:
        """Returns whether file has ReadUser permission flag."""
        return self.permissions() & core.QFileDevice.Permission.ReadUser

    def is_writable(self) -> bool:
        """Returns whether file has WriteUser permission flag."""
        return self.permissions() & core.QFileDevice.Permission.WriteUser

    def is_executable(self) -> bool:
        """Returns whether file has ExeUser permission flag."""
        return self.permissions() & core.QFileDevice.Permission.ExeUser


class FileDevice(FileDeviceMixin, core.QFileDevice):
    pass



from __future__ import annotations

from collections.abc import Iterator
import contextlib
import os
import pathlib

from typing import Literal

import qstylizer.parser
import qstylizer.style

from prettyqt import constants, core, gui
from prettyqt.utils import bidict, datatypes, get_repr


MarkdownFeatureStr = Literal["no_html", "commonmark", "github"]

MARKDOWN_FEATURES: bidict[MarkdownFeatureStr, gui.QTextDocument.MarkdownFeature] = bidict(
    no_html=gui.QTextDocument.MarkdownFeature.MarkdownNoHTML,
    commonmark=gui.QTextDocument.MarkdownFeature.MarkdownDialectCommonMark,
    github=gui.QTextDocument.MarkdownFeature.MarkdownDialectGitHub,
)

ResourceTypeStr = Literal["unknown", "html", "image", "stylesheet", "markdown", "user"]

RESOURCE_TYPES: bidict[ResourceTypeStr, gui.QTextDocument.ResourceType] = bidict(
    unknown=gui.QTextDocument.ResourceType.UnknownResource,
    html=gui.QTextDocument.ResourceType.HtmlResource,
    image=gui.QTextDocument.ResourceType.ImageResource,
    stylesheet=gui.QTextDocument.ResourceType.StyleSheetResource,
    markdown=gui.QTextDocument.ResourceType.MarkdownResource,
    user=gui.QTextDocument.ResourceType.UserResource,
)

StackStr = Literal["undo", "redo", "undo_and_redo"]

STACKS: bidict[StackStr, gui.QTextDocument.Stacks] = bidict(
    undo=gui.QTextDocument.Stacks.UndoStack,
    redo=gui.QTextDocument.Stacks.RedoStack,
    undo_and_redo=gui.QTextDocument.Stacks.UndoAndRedoStacks,
)

FindFlagStr = Literal["backward", "case_sensitive", "whole_words"]

FIND_FLAGS: bidict[FindFlagStr, gui.QTextDocument.FindFlag] = bidict(
    backward=gui.QTextDocument.FindFlag.FindBackward,
    case_sensitive=gui.QTextDocument.FindFlag.FindCaseSensitively,
    whole_words=gui.QTextDocument.FindFlag.FindWholeWords,
)

MetaInformationStr = Literal["document_title", "document_url", "css_media"]

META_INFORMATION: bidict[MetaInformationStr, gui.QTextDocument.MetaInformation] = bidict(
    document_title=gui.QTextDocument.MetaInformation.DocumentTitle,
    document_url=gui.QTextDocument.MetaInformation.DocumentUrl,
    css_media=gui.QTextDocument.MetaInformation.CssMedia,
)


class TextDocumentMixin(core.ObjectMixin):
    def __getitem__(self, index: int) -> gui.TextBlock:
        return gui.TextBlock(self.findBlockByNumber(index))

    def __len__(self) -> int:
        return self.blockCount()

    def __iter__(self) -> Iterator[gui.TextBlock]:
        return iter(
            gui.TextBlock(self.findBlockByNumber(i)) for i in range(self.blockCount())
        )

    def __repr__(self):
        return get_repr(self, self.toPlainText())

    def get_first_block(self) -> gui.TextBlock:
        return gui.TextBlock(self.firstBlock())

    def get_last_block(self) -> gui.TextBlock:
        return gui.TextBlock(self.lastBlock())

    def find_block_by_number(self, number: int) -> gui.TextBlock:
        # starts with 0-index
        block = self.findBlockByNumber(number)
        if not block.isValid():
            raise ValueError(
                f"{number} not a valid block index. Block count: {self.blockCount()}"
            )
        return gui.TextBlock(block)

    def find_block_by_line_number(self, line_number: int) -> gui.TextBlock:
        # starts with 0-index
        block = self.findBlockByLineNumber(line_number)
        if not block.isValid():
            raise ValueError(
                f"{line_number} not a valid line index. Line count: {self.lineCount()}"
            )
        return gui.TextBlock(block)

    def set_text(self, text: str):
        self.setPlainText(text)

    def get_base_url(self) -> core.Url:
        return core.Url(self.baseUrl())

    def get_default_font(self) -> gui.Font:
        return gui.Font(self.defaultFont())

    def set_default_text_option(self, opt: gui.QTextOption):
        self.setDefaultTextOption(gui.TextOption(opt))

    def get_default_text_option(self) -> gui.TextOption:
        return gui.TextOption(self.defaultTextOption())

    def clear_stacks(self, stack: StackStr | gui.QTextDocument.Stacks):
        """Clear undo / redo stack.

        Args:
            stack: stack to clear
        """
        self.clearUndoRedoStacks(STACKS.get_enum_value(stack))

    def set_default_cursor_move_style(
        self, style: constants.CursorMoveStyleStr | constants.CursorMoveStyle
    ):
        """Set the cursor move style.

        Args:
            style: cursor move style
        """
        self.setDefaultCursorMoveStyle(constants.CURSOR_MOVE_STYLE.get_enum_value(style))

    def get_default_cursor_move_style(self) -> constants.CursorMoveStyleStr:
        """Return current cursor move style.

        Returns:
            cursor move style
        """
        return constants.CURSOR_MOVE_STYLE.inverse[self.defaultCursorMoveStyle()]

    def set_meta_information(
        self, info: MetaInformationStr | gui.QTextDocument.MetaInformation, value: str
    ):
        """Set meta information.

        Args:
            info: meta information type
            value: value to set
        """
        self.setMetaInformation(META_INFORMATION.get_enum_value(info), value)

    def get_meta_information(
        self, info: MetaInformationStr | gui.QTextDocument.MetaInformation
    ) -> str:
        """Return specififed meta information.

        Args:
            info: meta information type

        Returns:
            meta information
        """
        return self.metaInformation(META_INFORMATION.get_enum_value(info))

    def add_resource(
        self,
        resource_type: ResourceTypeStr | gui.QTextDocument.ResourceType,
        name: datatypes.PathType,
        resource,
    ):
        url = core.Url(name)
        self.addResource(RESOURCE_TYPES.get_enum_value(resource_type), url, resource)

    @contextlib.contextmanager
    def edit_default_stylesheet(self) -> Iterator[qstylizer.style.StyleSheet]:
        ss = self.get_default_stylesheet()
        yield ss
        self.set_default_stylesheet(ss)

    def set_default_stylesheet(
        self, ss: None | qstylizer.style.StyleSheet | datatypes.PathType
    ):
        match ss:
            case os.PathLike():
                ss = pathlib.Path(ss).read_text()
            case None:
                ss = ""
            case qstylizer.style.StyleSheet():
                pass
            case _:
                raise TypeError(ss)
        self.setDefaultStyleSheet(str(ss))

    def get_default_stylesheet(self) -> qstylizer.style.StyleSheet:
        return qstylizer.parser.parse(self.defaultStyleSheet())

    def find_line_position(self, line_no: int) -> int:
        """Return index of the TextBlock corresponding to given line number."""
        lines = self.blockCount()
        assert 1 <= line_no <= lines
        return self.findBlockByLineNumber(line_no - 1).position()

    def write_to_file(
        self,
        path: datatypes.PathType,
        fmt: gui.textdocumentwriter.FormatStr | bytes | core.QByteArray = "plaintext",
    ) -> bool:
        writer = gui.TextDocumentWriter()
        writer.set_format(fmt)
        writer.set_file_name(path)
        return writer.write(self)

    def get_bytes(
        self, fmt: Literal["plaintext", "HTML", "markdown", "ODF"] = "ODF"
    ) -> bytes:
        return gui.TextDocumentWriter.serialize_document(self, fmt)

    def show_whitespace_and_tabs(self, show: bool):
        """Set show white spaces flag."""
        options = self.get_default_text_option()
        flag = gui.QTextOption.Flag.ShowTabsAndSpaces
        if show:
            options.setFlags(options.flags() | flag)
        else:
            options.setFlags(options.flags() & ~flag)
        self.setDefaultTextOption(options)

    def get_pixel_height(self, exact: bool = True) -> int:
        if exact:  # sourcery skip: extract-method
            layout = self.documentLayout()
            h = 0
            b = self.begin()
            while b != self.end():
                h += layout.blockBoundingRect(b).height()
                b = b.next()
            h += self.documentMargin()
            return h
        else:
            line_count = self.document().blockCount()
            font_metrics = gui.FontMetrics(self.defaultFont())
            return (max(line_count, 1) * font_metrics.height()) + self.documentMargin()


class TextDocument(TextDocumentMixin, gui.QTextDocument):
    pass


if __name__ == "__main__":
    doc = TextDocument("This is a test\nHello")
    doc.set_default_text_option(gui.QTextOption())
    a = doc.get_bytes()
    print(a)



from __future__ import annotations

from prettyqt import core
from prettyqt.qt import QtScxml


class ScxmlStateMachine(core.ObjectMixin, QtScxml.QScxmlStateMachine):
    """Interface to the state machines created from SCXML files."""



from __future__ import annotations

from typing import Literal

from prettyqt import constants, core, gui, syntaxhighlighters, widgets
from prettyqt.utils import bidict, classhelpers, colors, datatypes, texteditselecter


LINE_WRAP_MODE = bidict(
    none=widgets.QPlainTextEdit.LineWrapMode.NoWrap,
    widget_width=widgets.QPlainTextEdit.LineWrapMode.WidgetWidth,
)

LineWrapModeStr = Literal["none", "widget_width"]


class PlainTextEditMixin(widgets.AbstractScrollAreaMixin):
    value_changed = core.Signal(str)

    def __init__(self, text: str = "", **kwargs):
        super().__init__(**kwargs)
        self._allow_wheel_zoom = False
        self._hl = None
        self._current_line_color = gui.Color(0, 0, 0, 0)
        self._validation_color = gui.Color("orange")
        self.selecter = texteditselecter.TextEditSelecter(self)
        self.validator: gui.QValidator | None = None
        self.textChanged.connect(self._on_value_change)
        doc = gui.TextDocument(self)
        layout = widgets.PlainTextDocumentLayout(doc)
        doc.setDocumentLayout(layout)
        self.setDocument(doc)
        self.set_text(text)

    def __add__(self, other: str):
        self.append_text(other)
        return self

    def wheelEvent(self, event):
        """Handle wheel event for zooming."""
        if not self._allow_wheel_zoom:
            return super().wheelEvent(event)
        if event.modifiers() & constants.CTRL_MOD:
            self.zoomIn() if event.angleDelta().y() > 0 else self.zoomOut()
            event.accept()
        else:
            super().wheelEvent(event)

    def _get_map(self):
        maps = super()._get_map()
        maps |= {
            "lineWrapMode": LINE_WRAP_MODE,
            "wordWrapMode": gui.textoption.WORD_WRAP_MODE,
        }
        return maps

    def allow_wheel_zoom(self, do_zoom: bool = True):
        self._allow_wheel_zoom = do_zoom

    def append_text(
        self,
        text: str,
        newline: bool = True,
        ensure_visible: Literal["always", "when_bottom", "never"] = "always",
    ):
        scrollbar = self.verticalScrollBar()
        at_bottom = scrollbar.value() >= (scrollbar.maximum() - 4)
        prev_val = scrollbar.value()
        if newline:
            self.appendPlainText(text)
        else:
            self.selecter.move_cursor("end")
            self.insertPlainText(text)
            self.selecter.move_cursor("end")
        match ensure_visible:
            case "always":
                self.ensureCursorVisible()
            case "when_bottom":
                if at_bottom:
                    self.ensureCursorVisible()
            case "never":
                scrollbar.setValue(prev_val)

    def set_text(self, text: str):
        self.setPlainText(text)

    def set_syntaxhighlighter(
        self, syntax: str | gui.QSyntaxHighlighter, style: str | None = None
    ):
        if isinstance(syntax, gui.QSyntaxHighlighter):
            self._hl = syntax
        else:
            self._hl = syntaxhighlighters.PygmentsHighlighter(
                self.document(), syntax, style
            )

    def text(self) -> str:
        return self.toPlainText()

    def set_read_only(self, value: bool = True):
        """Make the PlainTextEdit read-only.

        Args:
            value: True, for read-only, otherwise False
        """
        self.setReadOnly(value)

    def show_whitespace_and_tabs(self, show: bool):
        """Set show white spaces flag."""
        doc = self.document()
        options = doc.defaultTextOption()
        flag = gui.QTextOption.Flag.ShowTabsAndSpaces
        if show:
            options.setFlags(options.flags() | flag)  # type: ignore
        else:
            options.setFlags(options.flags() & ~flag)  # type: ignore
        doc.setDefaultTextOption(options)

    def paintEvent(self, event: gui.QPaintEvent):
        if self._current_line_color:
            with gui.Painter(self.viewport()) as painter:
                cursor_rect = self.cursorRect()
                r = core.QRect(0, cursor_rect.top(), self.width(), cursor_rect.height())
                painter.set_pen(None)
                painter.setBrush(gui.Color(self._current_line_color))
                painter.drawRect(r)
        super().paintEvent(event)

    def set_word_wrap_mode(
        self, mode: gui.textoption.WordWrapModeStr | gui.QTextOption.WrapMode
    ):
        """Set word wrap mode.

        Args:
            mode: word wrap mode to use
        """
        self.setWordWrapMode(gui.textoption.WORD_WRAP_MODE.get_enum_value(mode))

    def get_word_wrap_mode(self) -> gui.textoption.WordWrapModeStr:
        """Get the current word wrap mode.

        Returns:
            Word wrap mode
        """
        return gui.textoption.WORD_WRAP_MODE.inverse[self.wordWrapMode()]

    def set_line_wrap_mode(
        self, mode: LineWrapModeStr | widgets.QPlainTextEdit.LineWrapMode
    ):
        """Set line wrap mode.

        Args:
            mode: line wrap mode to use
        """
        self.setLineWrapMode(LINE_WRAP_MODE.get_enum_value(mode))

    def get_line_wrap_mode(self) -> LineWrapModeStr:
        """Get the current wrap mode.

        Returns:
            Wrap mode
        """
        return LINE_WRAP_MODE.inverse[self.lineWrapMode()]

    def _on_value_change(self):
        self.value_changed.emit(self.text())
        if self.validator is not None:
            self._update_background()

    def _update_background(self):
        color = None if self.is_valid() else self._validation_color
        self.set_background_color(color)

    def set_validation_color(self, color: datatypes.ColorType):
        self._validation_color = colors.get_color(color).as_qt()

    def get_validation_color(self) -> gui.QColor:
        return self._validation_color

    def set_validator(
        self, validator: gui.QValidator | widgets.lineedit.ValidatorStr | None, **kwargs
    ) -> gui.QValidator:
        if isinstance(validator, str):
            ValidatorClass = classhelpers.get_class_for_id(gui.ValidatorMixin, validator)
            validator = ValidatorClass(**kwargs)
        self.validator = validator
        self._update_background()
        return validator

    def set_regex_validator(self, regex: str, flags=0) -> gui.RegularExpressionValidator:
        validator = gui.RegularExpressionValidator(self)
        validator.set_regex(regex, flags)
        self.set_validator(validator)
        return validator

    def is_valid(self) -> bool:
        if self.validator is None:
            return True
        val = self.validator.validate(self.text(), 0)
        return val[0] == self.validator.State.Acceptable

    def set_value(self, value: str):
        self.setPlainText(value)

    def get_value(self) -> str:
        return self.text()

    def set_current_line_color(self, color: datatypes.ColorType):
        if self._current_line_color is None:
            self.cursorPositionChanged.connect(self.selecter._update_on_block_change)
        if color is None:
            self.cursorPositionChanged.disconnect(self.selecter._update_on_block_change)
        self._current_line_color = colors.get_color(color) if color else None

    def get_current_line_color(self) -> gui.Color:
        return self._current_line_color

    current_line_color = core.Property(
        gui.QColor,
        get_current_line_color,
        set_current_line_color,
        doc="Color to use for current line highlighting",
    )

    def get_visible_line_span(self) -> tuple[int, int]:
        block = self.firstVisibleBlock()
        start = block.blockNumber()
        page_step = self.v_scrollbar.pageStep()
        end = min(start + page_step, self.blockCount())
        return (start, end)

    def get_pixel_height(self, exact: bool = True) -> int:
        """Returns the pixel height of the text area."""
        widget_margins = self.contentsMargins()
        doc_height = self.document().get_pixel_height(exact=exact)
        return widget_margins.top() + doc_height + widget_margins.bottom()

    validation_color = core.Property(
        gui.QColor,
        get_validation_color,
        set_validation_color,
        doc="Color to use for invalid text",
    )


class PlainTextEdit(PlainTextEditMixin, widgets.QPlainTextEdit):
    """Widget that is used to edit and display plain text."""


if __name__ == "__main__":
    from prettyqt import validators

    val = validators.RegexPatternValidator()
    app = widgets.app()
    widget = widgets.PlainTextEdit("\n".join(f"abc{i}" for i in range(1000)))
    widget.show_whitespace_and_tabs(True)
    # widget.set_validator(val)
    # with widget.current_cursor() as c:
    #     c.select_text(2, 4)
    widget.set_current_line_color(gui.Color(128, 128, 128, 30))
    widget.show()
    app.sleep(2)
    print(widget.get_visible_line_span())
    app.exec()



from __future__ import annotations

from prettyqt import gui


class AlphaNumericValidator(gui.Validator):
    """Validator which checks whether only alphanumeric characters are included."""

    ID = "alphanumeric"

    def __eq__(self, other: object):
        return isinstance(other, AlphaNumericValidator)

    def validate(self, text: str, pos: int = 0):
        if text.replace("_", "").isalnum():
            return self.State.Acceptable, text, pos
        return self.State.Invalid, text, pos


if __name__ == "__main__":
    from prettyqt import widgets

    val = AlphaNumericValidator()
    app = widgets.app()
    widget = widgets.LineEdit("Thisisatest")
    widget.setValidator(val)
    widget.show()
    app.exec()



from __future__ import annotations

from collections.abc import Iterator, MutableMapping

from typing_extensions import Self

from prettyqt.qt import QtCore
from prettyqt.utils import datatypes


class ProcessEnvironment(
    QtCore.QProcessEnvironment, MutableMapping, metaclass=datatypes.QABCMeta
):
    """Holds the environment variables that can be passed to a program."""

    def __bool__(self):
        return not self.isEmpty()

    def __contains__(self, other: str):
        return self.contains(other)

    def __getitem__(self, index: str) -> str:
        if index not in self:
            raise KeyError("Environment variable not set.")
        return self.value(index)

    def __delitem__(self, index: str):
        self.remove(index)

    def __setitem__(self, index: str, value: str):
        return self.insert(index, value)

    def __iter__(self) -> Iterator[str]:
        return iter(self.keys())

    def __len__(self):
        return len(self.keys())

    @classmethod
    def get_system_environment(cls) -> Self:
        return cls(cls.systemEnvironment())


if __name__ == "__main__":
    env = ProcessEnvironment.get_system_environment()
    env.update(dict(a="b"))
    print(list(env.items()))



from __future__ import annotations

from collections.abc import Sequence
import contextlib
import functools
import html
import os
import pathlib
import sys

from typing import TYPE_CHECKING, Any, Literal, overload

import qstylizer.parser
import qstylizer.style

from prettyqt import animations, constants, core, gui, iconprovider, widgets
from prettyqt.utils import colors, datatypes


if TYPE_CHECKING:
    from collections.abc import Callable, Iterator

    from prettyqt import custom_widgets


LayoutStr = Literal["horizontal", "vertical", "grid", "form", "stacked", "flow", "border"]

QWIDGETSIZE_MAX = 16_777_215  # widgets.QWIDGETSIZE_MAX

PositionPossibilityType = (
    Literal["parent", "window", "screen", "mouse"]
    | widgets.QWidget
    | core.QRect
    | core.QPoint
    | tuple[int, int]
    | tuple[int, int, int, int]
)


class WidgetMixin(core.ObjectMixin, gui.PaintDeviceMixin):
    def __init__(self, *args, margin: int | None = None, **kwargs):
        super().__init__(*args, **kwargs)
        self.fx = animations.Fx(self)
        if margin is not None:
            self.set_margin(margin)

    @classmethod
    def setup_example(cls):
        return cls()

    def _get_map(self):
        maps = super()._get_map()
        maps |= {
            "contextMenuPolicy": constants.CONTEXT_POLICY,
            "focusPolicy": constants.FOCUS_POLICY,
            "layoutDirection": constants.LAYOUT_DIRECTION,
            "windowModality": constants.WINDOW_MODALITY,
        }
        return maps

    def add_shortcut(
        self,
        keysequence: datatypes.KeyCombinationType,
        *,
        callback: Callable | None = None,
        context: constants.ShortcutContextStr = "window",
    ) -> gui.Shortcut:
        """Add shortcut to widget.

        Adds as shortcut for given callback and context to this widget.

        Args:
            keysequence: Key sequence
            callback: Callback for the shortcut
            context: context for this shortcut

        Returns:
            shortcut object
        """
        return gui.Shortcut(
            datatypes.to_keysequence(keysequence),
            self,
            callback,
            context=constants.SHORTCUT_CONTEXT.get_enum_value(context),
        )

    def get_win_id(self) -> int:
        return int(self.winId())

    def resize(self, *size):
        if isinstance(size[0], tuple):
            super().resize(*size[0])
        else:
            super().resize(*size)

    def set_width(self, width: int):
        self.resize(width, self.height())

    def set_height(self, height: int):
        self.resize(self.width(), height)

    def set_enabled(self, enabled: bool = True) -> None:
        self.setEnabled(enabled)

    def set_disabled(self) -> None:
        self.setEnabled(False)

    def insertAction(self, position_or_action: int | gui.QAction, action: gui.QAction):
        """Extend insertAction to also allow int index."""
        if isinstance(position_or_action, int):
            actions = self.actions()
            if len(actions) == 0 or position_or_action >= len(actions):
                # Insert as the first action or the last action.
                return super().addAction(action)
            position_or_action = actions[position_or_action]
        super().insertAction(position_or_action, action)

    def add_action(
        self,
        text: str | gui.Action,
        *,
        parent: widgets.QWidget | None = None,
        data: Any = None,
        **kwargs: datatypes.VariantType,
    ) -> gui.Action:
        """Add an action to the menu.

        Args:
            text: Label for the action
            parent: parent
            data: data for the Action
            kwargs: kwargs passed to action ctor
        Returns:
            Action added to menu
        """
        if isinstance(text, str):
            action = gui.Action(parent=parent or self, text=text, **kwargs)
        else:
            action = text
            action.setParent(self)
        self.addAction(action)
        action.setData(data)
        return action

    def add_actions(self, actions: Sequence[gui.QAction]):
        for i in actions:
            i.setParent(self)
        self.addActions(actions)

    def toggle_fullscreen(self) -> bool:
        """Toggle between fullscreen and regular size."""
        if self.isFullScreen():
            self.showNormal()
            return False
        else:
            self.showFullScreen()
            return True

    def toggle_maximized(self) -> bool:
        """Toggle between maximized and regular size."""
        if self.isMaximized():
            self.showNormal()
            return False
        else:
            self.showMaximized()
            return True

    def map_to(
        self,
        widget: widgets.QWidget | Literal["global", "parent", "window"],
        pos_or_rect: datatypes.PointType
        | datatypes.RectType
        | datatypes.PointFType
        | datatypes.RectFType,
    ) -> core.QRect | core.QRectF | core.QPoint | core.QPointF:
        """Map a point or rect to a widget, global position or parent.

        Arguments:
            widget: What to map to.
            pos_or_rect: Point or rect to map.
        """
        match pos_or_rect:
            case int(), int():
                pos_or_rect = core.QPoint(*pos_or_rect)
            case float(), float():
                pos_or_rect = core.QPointF(*pos_or_rect)
            case int(), int(), int(), int():
                pos_or_rect = core.QRect(*pos_or_rect)
            case float(), float(), float(), float():
                pos_or_rect = core.QRectF(*pos_or_rect)
        match pos_or_rect, widget:
            case core.QRect() | core.QRectF(), widgets.QWidget():
                top_left = super().mapTo(widget, pos_or_rect.topLeft())
                bottom_right = super().mapTo(widget, pos_or_rect.bottomRight())
                return type(pos_or_rect)(top_left, bottom_right)
            case core.QPoint() | core.QPointF(), widgets.QWidget():
                return super().mapTo(widget, pos_or_rect)
            case core.QRect() | core.QRectF(), "parent":
                top_left = super().mapToParent(pos_or_rect.topLeft())
                bottom_right = super().mapToParent(pos_or_rect.bottomRight())
                return type(pos_or_rect)(top_left, bottom_right)
            case core.QPoint() | core.QPointF(), "parent":
                return super().mapToParent(pos_or_rect)
            case core.QRect() | core.QRectF(), "window":
                top_left = super().mapTo(self.window(), pos_or_rect.topLeft())
                bottom_right = super().mapTo(self.window(), pos_or_rect.bottomRight())
                return type(pos_or_rect)(top_left, bottom_right)
            case core.QPoint() | core.QPointF(), "window":
                return super().mapTo(self.window(), pos_or_rect)
            case core.QRect() | core.QRectF(), "global":
                top_left = super().mapToGlobal(pos_or_rect.topLeft())
                bottom_right = super().mapToGlobal(pos_or_rect.bottomRight())
                return type(pos_or_rect)(top_left, bottom_right)
            case core.QPoint() | core.QPointF(), "global":
                return super().mapToGlobal(pos_or_rect)
            case _:
                raise ValueError(pos_or_rect)

    def map_from(
        self,
        widget: widgets.QWidget | Literal["global", "parent", "window"],
        pos_or_rect: datatypes.PointType
        | datatypes.RectType
        | datatypes.PointFType
        | datatypes.RectFType,
    ) -> core.QRect | core.QRectF | core.QPoint | core.QPointF:
        """Map a point or rect from a widget, global position or parent.

        Arguments:
            widget: What to map from.
            pos_or_rect: Point or rect to map.
        """
        match pos_or_rect:
            case int(), int():
                pos_or_rect = core.QPoint(*pos_or_rect)
            case float(), float():
                pos_or_rect = core.QPointF(*pos_or_rect)
            case int(), int(), int(), int():
                pos_or_rect = core.QRect(*pos_or_rect)
            case float(), float(), float(), float():
                pos_or_rect = core.QRectF(*pos_or_rect)
        match pos_or_rect, widget:
            case core.QRect() | core.QRectF(), widgets.QWidget():
                top_left = super().mapFrom(widget, pos_or_rect.topLeft())
                bottom_right = super().mapFrom(widget, pos_or_rect.bottomRight())
                return type(pos_or_rect)(top_left, bottom_right)
            case core.QPoint() | core.QPointF(), widgets.QWidget():
                return super().mapFrom(widget, pos_or_rect)
            case core.QRect() | core.QRectF(), "parent":
                top_left = super().mapFromParent(pos_or_rect.topLeft())
                bottom_right = super().mapFromParent(pos_or_rect.bottomRight())
                return type(pos_or_rect)(top_left, bottom_right)
            case core.QPoint() | core.QPointF(), "parent":
                return super().mapFromParent(pos_or_rect)
            case core.QRect() | core.QRectF(), "window":
                top_left = super().mapFrom(self.window(), pos_or_rect.topLeft())
                bottom_right = super().mapFrom(self.window(), pos_or_rect.bottomRight())
                return type(pos_or_rect)(top_left, bottom_right)
            case core.QPoint() | core.QPointF(), "window":
                return super().mapFrom(self.window(), pos_or_rect)
            case core.QRect() | core.QRectF(), "global":
                top_left = super().mapFromGlobal(pos_or_rect.topLeft())
                bottom_right = super().mapFromGlobal(pos_or_rect.bottomRight())
                return type(pos_or_rect)(top_left, bottom_right)
            case core.QPoint() | core.QPointF(), "global":
                return super().mapFromGlobal(pos_or_rect)
            case _:
                raise ValueError(pos_or_rect)

    def raise_to_top(self):
        """Brings a widget to top with all means available."""
        if sys.platform.startswith("win"):
            from prettyqt.utils.platforms.windows import misc

            misc.raise_to_top()
        # state = (self.windowState() & ~Qt.WindowMinimized) | Qt.WindowActive
        # self.setWindowState(state)
        self.raise_()
        self.show()
        self.activateWindow()

    def set_icon(self, icon: datatypes.IconType):
        """Set the window icon.

        Args:
            icon: icon to use
        """
        color = widgets.app().get_window_icon_color()
        icon = iconprovider.get_icon(icon, color=color)
        super().setWindowIcon(icon)

    setWindowIcon = set_icon

    def get_icon(self) -> gui.Icon | None:
        """Get the window icon (returns None if not existing)."""
        icon = super().windowIcon()
        return None if icon.isNull() else gui.Icon(icon)

    def show_tooltip(self, duration: int | None = None):
        """Show the tooltip of this widget for given time.

        When no duration is given, it will get calculated based on length.
        """
        if duration is None:
            duration = -1  # automatic
        pos = self.map_to("global", (0, 0))
        widgets.ToolTip.showText(pos, self.toolTip(), msecShowTime=duration)

    # lets be gentle and allow all reasonable signatures for the size setters.
    # tuples as well as passing two args is possible.
    @functools.singledispatchmethod
    def set_min_size(self, size: core.QSize | tuple[int | None, int | None]):
        match size:
            case int() | None as x, int() | None as y:
                super().setMinimumSize(x or 0, y or 0)
            case core.QSize():
                super().setMinimumSize(size)
            case _:
                raise TypeError(size)

    setMinimumSize = set_min_size

    @set_min_size.register
    def _(self, x: int, y: int | None):
        self.set_min_size((x, y))

    @set_min_size.register  # these can be merged when min py version is 3.11
    def _(self, x: None, y: int | None):
        self.set_min_size((x, y))

    @functools.singledispatchmethod
    def set_max_size(self, size: core.QSize | tuple[int | None, int | None]):
        match size:
            case int() | None as x, int() | None as y:
                x = QWIDGETSIZE_MAX if x is None else x
                y = QWIDGETSIZE_MAX if y is None else y
                super().setMaximumSize(x, y)
            case _:
                super().setMaximumSize(size)

    setMaximumSize = set_min_size

    @set_max_size.register
    def _(self, x: int, y: int | None):
        self.set_max_size((x, y))

    @set_max_size.register  # these can be merged when min py version is 3.11
    def _(self, x: None, y: int | None):
        self.set_max_size((x, y))

    def set_min_width(self, width: int | None):
        super().setMinimumWidth(width or 0)

    setMinimumWidth = set_min_width

    def set_max_width(self, width: int | None):
        if width is None:
            width = QWIDGETSIZE_MAX
        super().setMaximumWidth(width)

    setMaximumWidth = set_max_width

    def set_min_height(self, height: int | None):
        super().setMinimumHeight(height or 0)

    setMinimumHeight = set_min_height

    def set_max_height(self, height: int | None):
        if height is None:
            height = QWIDGETSIZE_MAX
        super().setMaximumHeight(height)

    setMaximumHeight = set_max_height

    def setWindowTitle(self, title: str):
        if not self.objectName() and widgets.app().is_debug():
            self.setObjectName(title)
        super().setWindowTitle(title)

    def set_title(self, title: str):
        self.setWindowTitle(title)

    def get_title(self) -> str:
        return self.windowTitle()

    def set_tooltip(
        self,
        tooltip: str | datatypes.PathType,
        *,
        size: datatypes.SizeType | None = None,
        rich_text: bool = False,
    ):
        """Set a tooltip for this widget.

        In image can get displayed by passing a PathLike object.
        """
        if isinstance(tooltip, os.PathLike):
            path = os.fspath(tooltip)
            if size is None:
                tooltip = f"<img src={path!r}>"
            else:
                if isinstance(size, core.QSize):
                    size = (size.width(), size.height())
                tooltip = f'<img src={path!r} width="{size[0]}" height="{size[1]}">'
        tooltip = tooltip.replace("\n", "<br/>")
        if rich_text:
            tooltip = f"<html>{html.escape(tooltip)}</html>"
        super().setToolTip(tooltip)

    def set_font(
        self,
        font_name: gui.QFont | str | None = None,
        *,
        font_size: int | None = None,
        weight: int | None = None,
        italic: bool = False,
    ) -> gui.QFont:
        """Set the font for this widget."""
        if isinstance(font_name, gui.QFont):
            super().setFont(font_name)
            return font_name
        if font_size is None:
            font_size = -1
        if weight is None:
            weight = -1
        if font_name is None:
            font_name = self.font().family()
        font = gui.Font(font_name, font_size, weight, italic)
        super().setFont(font)
        return font

    def get_font(self) -> gui.Font:
        return gui.Font(self.font())

    def get_foreground_role(self) -> gui.palette.RoleStr:
        """Set foreground role for this widget."""
        return gui.palette.ROLE.inverse[self.foregroundRole()]

    def set_foreground_role(self, role: gui.palette.RoleStr | gui.Palette.ColorRole):
        """Set foreground role for this widget."""
        self.setForegroundRole(gui.palette.ROLE.get_enum_value(role))

    def get_background_role(self) -> gui.palette.RoleStr:
        """Get background role for this widget."""
        return gui.palette.ROLE.inverse[self.backgroundRole()]

    def set_background_role(self, role: gui.palette.RoleStr | gui.Palette.ColorRole):
        """Get foreground role for this widget."""
        self.setBackgroundRole(gui.palette.ROLE.get_enum_value(role))

    def set_window_flags(self, *flags: constants.WindowTypeStr, append: bool = False):
        result = constants.WINDOW_TYPE.merge_flags(flags)
        if append:
            result = result | self.windowFlags()
        self.setWindowFlags(result)

    def set_flags(
        self,
        *,
        minimize: bool | None = None,
        maximize: bool | None = None,
        close: bool | None = None,
        stay_on_top: bool | None = None,
        frameless: bool | None = None,
        window: bool | None = None,
        dialog: bool | None = None,
        tooltip: bool | None = None,
        tool: bool | None = None,
        customize: bool | None = None,
        cover: bool | None = None,
        window_title: bool | None = None,
    ):
        """Set window flags.

        Arguments:
            minimize: set WindowMinimizeButtonHint flag
            maximize: set WindowMaximizeButtonHint flag
            close: set WindowCloseButtonHint flag
            stay_on_top: set WindowStaysOnTopHint flag
            frameless: set FramelessWindowHint flag
            window: set Window flag
            dialog: set Dialog flag
            tooltip: set ToolTip flag
            tool: set Tool flag
            customize: set CustomizeWindowHint flag
            cover: set CoverWindow flag
            window_title: set WindowTitleHint flag
        """
        flags = {
            constants.WindowType.WindowMinimizeButtonHint: minimize,
            constants.WindowType.WindowMaximizeButtonHint: maximize,
            constants.WindowType.WindowCloseButtonHint: close,
            constants.WindowType.WindowStaysOnTopHint: stay_on_top,
            constants.WindowType.FramelessWindowHint: frameless,
            constants.WindowType.Window: window,
            constants.WindowType.Dialog: dialog,
            constants.WindowType.ToolTip: tooltip,
            constants.WindowType.Tool: tool,
            constants.WindowType.CustomizeWindowHint: customize,
            constants.WindowType.CoverWindow: cover,
            constants.WindowType.WindowTitleHint: window_title,
        }
        for k, v in flags.items():
            if v is not None:
                self.setWindowFlag(k, v)

    def set_attributes(
        self,
        *,
        accept_drops: bool | None = None,
        always_show_tooltips: bool | None = None,
        custom_whats_this: bool | None = None,
        delete_on_close: bool | None = None,
        disabled: bool | None = None,
        dont_show_on_screen: bool | None = None,
        force_disabled: bool | None = None,
        force_updates_disabled: bool | None = None,
        hover: bool | None = None,
        input_method_enabled: bool | None = None,
        keyboard_focus_change: bool | None = None,
        key_compression: bool | None = None,
        layout_on_entire_rect: bool | None = None,
        layout_uses_widget_rect: bool | None = None,
        mapped: bool | None = None,
        mouse_no_mask: bool | None = None,
        mouse_tracking: bool | None = None,
        moved: bool | None = None,
        no_child_events_for_parent: bool | None = None,
        no_child_events_from_children: bool | None = None,
        no_mouse_replay: bool | None = None,
        no_mouse_propagation: bool | None = None,
        transparent_for_mouse_events: bool | None = None,
        no_system_background: bool | None = None,
        opaque_paint_event: bool | None = None,
        outside_ws_range: bool | None = None,
        paint_on_screen: bool | None = None,
        paint_unclipped: bool | None = None,
        pending_move_event: bool | None = None,
        pending_resize_egent: bool | None = None,
        quit_on_close: bool | None = None,
        resized: bool | None = None,
        right_to_left: bool | None = None,
        set_cursor: bool | None = None,
        set_font: bool | None = None,
        set_palette: bool | None = None,
        set_style: bool | None = None,
        static_contents: bool | None = None,
        style_sheet: bool | None = None,
        style_sheet_target: bool | None = None,
        tablet_tracking: bool | None = None,
        translucent_background: bool | None = None,
        under_mouse: bool | None = None,
        updates_disabled: bool | None = None,
        window_modified: bool | None = None,
        window_propagation: bool | None = None,
        mac_always_show_tool_window: bool | None = None,
        set_locale: bool | None = None,
        styled_background: bool | None = None,
        show_without_activating: bool | None = None,
        native_window: bool | None = None,
        dont_create_native_ancestors: bool | None = None,
        accept_touch_events: bool | None = None,
        touch_pad_single_touch_events: bool | None = None,
        always_stack_on_top: bool | None = None,
        contents_margins_respects_safe_area: bool | None = None,
    ):
        """Set window attributes.

        Attributes:
            accept_drops: set AcceptDrops attribute
            always_show_tooltips: set AlwaysShowToolTips attribute
            custom_whats_this: set CustomWhatsThis attribute
            delete_on_close: set DeleteOnClose attribute
            disabled: set Disabled attribute
            dont_show_on_screen: set DontShowOnScreen attribute
            force_disabled: set ForceDisabled attribute
            force_updates_disabled: set ForceUpdatesDisabled attribute
            hover: set Hover attribute
            input_method_enabled: set InputMethodEnabled attribute
            keyboard_focus_change: set KeyboardFocusChange attribute
            key_compression: set KeyCompression attribute
            layout_on_entire_rect: set LayoutOnEntireRect attribute
            layout_uses_widget_rect: set LayoutUsesWidgetRect attribute
            mapped: set Mapped attribute
            mouse_no_mask: set MouseNoMask attribute
            mouse_tracking: set MouseTracking attribute
            moved: set Moved attribute
            no_child_events_for_parent: set NoChildEventsForParent attribute
            no_child_events_from_children: set NoChildEventsFromChildren attribute
            no_mouse_replay: set NoMouseReplay attribute
            no_mouse_propagation: set NoMousePropagation attribute
            transparent_for_mouse_events: set TransparentForMouseEvents attribute
            no_system_background: set NoSystemBackground attribute
            opaque_paint_event: set OpaquePaintEvent attribute
            outside_ws_range: set OutsideWSRange attribute
            paint_on_screen: set PaintOnScreen attribute
            paint_unclipped: set PaintUnclipped attribute
            pending_move_event: set PendingMoveEvent attribute
            pending_resize_egent: set PendingResizeEvent attribute
            quit_on_close: set QuitOnClose attribute
            resized: set Resized attribute
            right_to_left: set RightToLeft attribute
            set_cursor: set SetCursor attribute
            set_font: set SetFont attribute
            set_palette: set SetPalette attribute
            set_style: set SetStyle attribute
            static_contents: set StaticContents attribute
            style_sheet: set StyleSheet attribute
            style_sheet_target: set StyleSheetTarget attribute
            tablet_tracking: set TabletTracking attribute
            translucent_background: set TranslucentBackground attribute
            under_mouse: set UnderMouse attribute
            updates_disabled: set UpdatesDisabled attribute
            window_modified: set WindowModified attribute
            window_propagation: set WindowPropagation attribute
            mac_always_show_tool_window: set MacAlwaysShowToolWindow attribute
            set_locale: set SetLocale attribute
            styled_background: set StyledBackground attribute
            show_without_activating: set ShowWithoutActivating attribute
            native_window: set NativeWindow attribute
            dont_create_native_ancestors: set DontCreateNativeAncestors attribute
            accept_touch_events: set AcceptTouchEvents attribute
            touch_pad_single_touch_events: set TouchPadAcceptSingleTouchEvents attribute
            always_stack_on_top: set AlwaysStackOnTop attribute
            contents_margins_respects_safe_area: set ContentsMarginsRespectsSafeArea
                                                 attribute
        """
        Attr = constants.WidgetAttribute
        flags = {
            Attr.WA_AcceptDrops: accept_drops,
            Attr.WA_AlwaysShowToolTips: always_show_tooltips,
            Attr.WA_CustomWhatsThis: custom_whats_this,
            Attr.WA_DeleteOnClose: delete_on_close,
            Attr.WA_Disabled: disabled,
            Attr.WA_DontShowOnScreen: dont_show_on_screen,
            Attr.WA_ForceDisabled: force_disabled,
            Attr.WA_ForceUpdatesDisabled: force_updates_disabled,
            Attr.WA_Hover: hover,
            Attr.WA_InputMethodEnabled: input_method_enabled,
            Attr.WA_KeyboardFocusChange: keyboard_focus_change,
            Attr.WA_KeyCompression: key_compression,
            Attr.WA_LayoutOnEntireRect: layout_on_entire_rect,
            Attr.WA_LayoutUsesWidgetRect: layout_uses_widget_rect,
            Attr.WA_Mapped: mapped,
            Attr.WA_MouseNoMask: mouse_no_mask,
            Attr.WA_MouseTracking: mouse_tracking,
            Attr.WA_Moved: moved,
            Attr.WA_NoChildEventsForParent: no_child_events_for_parent,
            Attr.WA_NoChildEventsFromChildren: no_child_events_from_children,
            Attr.WA_NoMouseReplay: no_mouse_replay,
            Attr.WA_NoMousePropagation: no_mouse_propagation,
            Attr.WA_TransparentForMouseEvents: transparent_for_mouse_events,
            Attr.WA_NoSystemBackground: no_system_background,
            Attr.WA_OpaquePaintEvent: opaque_paint_event,
            Attr.WA_OutsideWSRange: outside_ws_range,
            Attr.WA_PaintOnScreen: paint_on_screen,
            Attr.WA_PaintUnclipped: paint_unclipped,
            Attr.WA_PendingMoveEvent: pending_move_event,
            Attr.WA_PendingResizeEvent: pending_resize_egent,
            Attr.WA_QuitOnClose: quit_on_close,
            Attr.WA_Resized: resized,
            Attr.WA_RightToLeft: right_to_left,
            Attr.WA_SetCursor: set_cursor,
            Attr.WA_SetFont: set_font,
            Attr.WA_SetPalette: set_palette,
            Attr.WA_SetStyle: set_style,
            Attr.WA_StaticContents: static_contents,
            Attr.WA_StyleSheet: style_sheet,
            Attr.WA_StyleSheetTarget: style_sheet_target,
            Attr.WA_TabletTracking: tablet_tracking,
            Attr.WA_TranslucentBackground: translucent_background,
            Attr.WA_UnderMouse: under_mouse,
            Attr.WA_UpdatesDisabled: updates_disabled,
            Attr.WA_WindowModified: window_modified,
            Attr.WA_WindowPropagation: window_propagation,
            Attr.WA_MacAlwaysShowToolWindow: mac_always_show_tool_window,
            Attr.WA_SetLocale: set_locale,
            Attr.WA_StyledBackground: styled_background,
            Attr.WA_ShowWithoutActivating: show_without_activating,
            Attr.WA_NativeWindow: native_window,
            Attr.WA_DontCreateNativeAncestors: dont_create_native_ancestors,
            Attr.WA_AcceptTouchEvents: accept_touch_events,
            Attr.WA_TouchPadAcceptSingleTouchEvents: touch_pad_single_touch_events,
            Attr.WA_AlwaysStackOnTop: always_stack_on_top,
            Attr.WA_ContentsMarginsRespectsSafeArea: contents_margins_respects_safe_area,
        }
        for k, v in flags.items():
            if v is not None:
                self.setAttribute(k, v)

    def set_modality(
        self, modality: constants.WindowModalityStr | constants.WindowModality
    ):
        """Set modality for the dialog.

        Args:
            modality: modality for the main window
        """
        self.setWindowModality(constants.WINDOW_MODALITY.get_enum_value(modality))

    def get_modality(self) -> constants.WindowModalityStr:
        """Get the current modality modes as a string.

        Returns:
            modality mode
        """
        return constants.WINDOW_MODALITY.inverse[self.windowModality()]

    def set_size_policy(
        self,
        horizontal: widgets.sizepolicy.SizePolicyStr
        | widgets.QSizePolicy.Policy
        | None = None,
        vertical: widgets.sizepolicy.SizePolicyStr
        | widgets.QSizePolicy.Policy
        | None = None,
    ):
        """Set the size policy.

        Args:
            horizontal: horizontal size policy
            vertical: vertical size policy
        """
        sp = self.get_size_policy()
        if horizontal is not None:
            sp.set_horizontal_policy(horizontal)
        if vertical is not None:
            sp.set_vertical_policy(vertical)
        self.setSizePolicy(sp)

    def get_size_policy(self) -> widgets.SizePolicy:
        """Get size policy."""
        qpol = self.sizePolicy()
        if isinstance(qpol, widgets.SizePolicy):
            return qpol
        return widgets.SizePolicy.clone(qpol)

    def get_palette(self) -> gui.Palette:
        return gui.Palette(self.palette())

    def set_background_color(self, color: datatypes.ColorType):
        col_str = "" if color is None else colors.get_color(color).name()
        with self.edit_stylesheet() as ss:
            ss.backgroundColor.setValue(col_str)

    @contextlib.contextmanager
    def grab_mouse_events(
        self, cursor_shape: constants.CursorShapeStr | None = None
    ) -> Iterator[None]:
        """Context manager to grab mouse events."""
        if cursor_shape is not None:
            self.grabMouse(constants.CURSOR_SHAPE[cursor_shape])
        else:
            self.grabMouse()
        yield None
        self.releaseMouse()

    @contextlib.contextmanager
    def grab_keyboard_events(self) -> Iterator[None]:
        """Context manager to grab keyboard events."""
        self.grabKeyboard()
        yield None
        self.releaseKeyboard()

    @contextlib.contextmanager
    def updates_off(self) -> Iterator[None]:
        """Context manager to turn off updates for this widget."""
        updates = self.updatesEnabled()
        self.setUpdatesEnabled(False)
        yield None
        self.setUpdatesEnabled(updates)

    @contextlib.contextmanager
    def edit_stylesheet(self) -> Iterator[qstylizer.style.StyleSheet]:
        """Context manager to edit the stylesheet (using qstylizer)."""
        ss = self.get_stylesheet()
        yield ss
        self.set_stylesheet(ss)

    def set_stylesheet(
        self, ss: None | str | qstylizer.style.StyleSheet | datatypes.PathType
    ):
        """Set stylesheet for this widget."""
        match ss:
            case None:
                ss = ""
            case str():
                pass
            case os.PathLike():
                ss = pathlib.Path(ss).read_text()
            case qstylizer.style.StyleSheet():
                ss = str(ss)
        self.setStyleSheet(ss)

    def get_stylesheet(self) -> qstylizer.style.StyleSheet:
        """Get current stylesheet."""
        try:
            return qstylizer.parser.parse(self.styleSheet())
        except ValueError:
            return qstylizer.style.StyleSheet()

    @contextlib.contextmanager
    def edit_palette(self) -> Iterator[gui.Palette]:
        """Context manager to edit the palette of the widget."""
        palette = gui.Palette(self.palette())
        yield palette
        self.setPalette(palette)

    @contextlib.contextmanager
    def edit_font(self) -> Iterator[gui.Font]:
        """Context manager to edit the font of the widget."""
        font = gui.Font(self.font())
        yield font
        self.setFont(font)

    def set_context_menu_policy(
        self, policy: constants.ContextPolicyStr | constants.ContextMenuPolicy
    ):
        """Set contextmenu policy for given item view.

        Args:
            policy: contextmenu policy to use
        """
        self.setContextMenuPolicy(constants.CONTEXT_POLICY.get_enum_value(policy))

    def get_context_menu_policy(self) -> constants.ContextPolicyStr:
        """Return current contextmenu policy.

        Returns:
            contextmenu policy
        """
        return constants.CONTEXT_POLICY.inverse[self.contextMenuPolicy()]

    def set_window_state(self, state: constants.WindowStateStr | constants.WindowState):
        """Set window state for given item view.

        Args:
            state: window state to use
        """
        self.setWindowState(constants.WINDOW_STATES.get_enum_value(state))

    def get_window_state(self) -> constants.WindowStateStr:
        """Return current window state.

        Returns:
            window state
        """
        return constants.WINDOW_STATES.inverse[self.windowState()]

    def set_custom_menu(self, method: Callable):
        self.set_context_menu_policy("custom")
        self.customContextMenuRequested.connect(method)

    @property
    def box(self) -> widgets.Layout:
        return self.layout()

    @box.setter
    def box(self, layout: widgets.Layout):
        self.set_layout(layout)

    @overload
    def set_layout(
        self, layout: Literal["horizontal"], margin: int | None = None, **kwargs
    ) -> widgets.HBoxLayout:
        pass

    @overload
    def set_layout(
        self, layout: Literal["vertical"], margin: int | None = None, **kwargs
    ) -> widgets.VBoxLayout:
        pass

    @overload
    def set_layout(
        self, layout: Literal["grid"], margin: int | None = None, **kwargs
    ) -> widgets.GridLayout:
        pass

    @overload
    def set_layout(
        self, layout: Literal["border"], margin: int | None = None, **kwargs
    ) -> custom_widgets.BorderLayout:
        pass

    @overload
    def set_layout(
        self, layout: Literal["flow"], margin: int | None = None, **kwargs
    ) -> custom_widgets.FlowLayout:
        pass

    @overload
    def set_layout(
        self, layout: Literal["form"], margin: int | None = None, **kwargs
    ) -> widgets.FormLayout:
        pass

    @overload
    def set_layout(
        self, layout: Literal["stacked"], margin: int | None = None, **kwargs
    ) -> widgets.StackedLayout:
        pass

    def set_layout(
        self,
        layout: LayoutStr | widgets.QLayout,
        *,
        margin: int | None = None,
        **kwargs: datatypes.VariantType,
    ) -> widgets.QLayout:
        """Quick way to set a layout.

        Sets layout to given layout, also allows setting margin and spacing.

        Args:
            layout: Layout to set
            margin: margin to use in pixels
            kwargs: keyword arguments passed to layout

        Returns:
            Layout
        """
        from prettyqt import custom_widgets

        match layout:
            case "horizontal":
                layout = widgets.HBoxLayout(**kwargs)
            case "vertical":
                layout = widgets.VBoxLayout(**kwargs)
            case "grid":
                layout = widgets.GridLayout(**kwargs)
            case "form":
                layout = widgets.FormLayout(**kwargs)
            case "stacked":
                layout = widgets.StackedLayout(**kwargs)
            case "flow":
                layout = custom_widgets.FlowLayout(**kwargs)
            case "border":
                layout = custom_widgets.BorderLayout(**kwargs)
            case widgets.QLayout():
                layout = layout
            case _:
                raise ValueError(f"Invalid Layout {layout}")
        self.setLayout(layout)
        if margin is not None:
            layout.set_margin(margin)
        return layout

    def position_on(
        self,
        where: PositionPossibilityType,
        how: Literal[
            "center",
            "top",
            "left",
            "bottom",
            "right",
            "top_left",
            "top_right",
            "bottom_left",
            "bottom_right",
        ] = "center",
        *,
        scale_ratio: int | None = None,
        x_offset: int = 0,
        y_offset: int = 0,
    ):
        """Position widget on another widget / window / screen.

        Arguments:
            where: where to positin on
            how: How to align
            scale_ratio: Resize to scale_ratio * target size
            x_offset: additional x offset for final position
            y_offset: additional y offset for final position
        """
        do_scale = True
        match where:
            case "mouse":
                geom = core.Rect(gui.Cursor.pos(), gui.Cursor.pos())
                do_scale = False
            case core.QPoint():
                geom = core.Rect(where, where)
                do_scale = False
            case (int(), int()):
                p = core.Point(*where)
                geom = core.Rect(p, p)
                do_scale = False
            case (int(), int(), int(), int()):
                geom = core.Rect(*where)
            case "parent":
                geom = self.parent().frameGeometry()
            case "window":
                geom = self.window().frameGeometry()
            case widgets.QWidget():
                geom = where.frameGeometry()
            case core.QRect():
                geom = where
            case "screen":
                geom = gui.GuiApplication.primaryScreen().geometry()
            case _:
                raise ValueError(where)
        if scale_ratio is not None and do_scale:
            self.resize(
                int(geom.width() * scale_ratio),
                int(geom.height() * scale_ratio),
            )
        own_geo = self.frameGeometry()
        match how:
            case "center":
                new = geom.center()
            case "top":
                new = core.Point(geom.center().x(), geom.top() + own_geo.height() // 2)
            case "bottom":
                new = core.Point(geom.center().x(), geom.bottom() - own_geo.height() // 2)
            case "left":
                new = core.Point(geom.left() + own_geo.width() // 2, geom.center().y())
            case "right":
                new = core.Point(geom.right() - own_geo.width() // 2, geom.center().y())
            case "top_right":
                new = core.Point(
                    geom.right() - own_geo.width() // 2,
                    geom.top() + own_geo.height() // 2,
                )
            case "top_left":
                new = core.Point(
                    geom.left() + own_geo.width() // 2,
                    geom.top() + own_geo.height() // 2,
                )
            case "bottom_right":
                new = core.Point(
                    geom.right() - own_geo.width() // 2,
                    geom.bottom() - own_geo.height() // 2,
                )
            case "bottom_left":
                new = core.Point(
                    geom.left() + own_geo.width() // 2,
                    geom.bottom() - own_geo.height() // 2,
                )
            case _:
                raise ValueError(how)
        new = core.Point(new.x() + x_offset, new.y() + y_offset)
        own_geo.moveCenter(new)
        self.move(own_geo.topLeft())

    def set_cursor(
        self, cursor: constants.CursorShapeStr | constants.CursorShape | gui.QCursor
    ):
        """Set the cursor for this widget."""
        if isinstance(cursor, gui.QCursor):
            curs = cursor
        else:
            curs = gui.Cursor(constants.CURSOR_SHAPE.get_enum_value(cursor))
        self.setCursor(curs)

    def set_focus_policy(self, policy: constants.FocusPolicyStr | constants.FocusPolicy):
        """Set the way the widget accepts keyboard focus.

        Args:
            policy (str): Focus policy
        """
        self.setFocusPolicy(constants.FOCUS_POLICY.get_enum_value(policy))

    def get_focus_policy(self) -> constants.FocusPolicyStr:
        """Return waay the widget accepts keyboard focus.

        Returns:
            str: Focus policy
        """
        return constants.FOCUS_POLICY.inverse[self.focusPolicy()]

    def set_focus(self, reason: constants.FocusReasonStr | None = None):
        if reason is None:
            self.setFocus()
        else:
            self.setFocus(constants.FOCUS_REASONS[reason])

    def set_font_size(self, size: int):
        font = self.font()
        font.setPointSize(size)
        self.setFont(font)

    def get_font_metrics(self) -> gui.FontMetrics:
        return gui.FontMetrics(self.fontMetrics())

    def get_font_info(self) -> gui.FontInfo:
        return gui.FontInfo(self.fontInfo())

    def set_margin(self, margin: datatypes.MarginsType):
        """Set content margins for the widget.

        Arguments:
            margin: margins to use
        """
        self.setContentsMargins(datatypes.to_margins(margin))

    def raise_dock(self) -> bool:
        w = self.find_parent(widgets.QDockWidget)
        if w is None:
            return False
        w.setVisible(True)
        w.raise_()
        return True

    def set_mask(
        self,
        area: datatypes.RectType | gui.QRegion | gui.QBitmap | None,
        typ: gui.region.RegionTypeStr = "rectangle",
    ):
        """Set mask of the widget.

        Arguments:
            area: Mask area
            typ: type of region (only used if area is a QRegion)
        """
        match area:
            case None:
                self.clearMask()
                return
            case (int(), int(), int(), int()):
                area = gui.Region(*area, gui.region.REGION_TYPE[typ])
            case core.QRect():
                area = gui.Region(area, gui.region.REGION_TYPE[typ])
        self.setMask(area)

    def set_window_file_path(self, path: datatypes.PathType):
        self.setWindowFilePath(os.fspath(path))

    def get_window_file_path(self) -> pathlib.Path | None:
        path = self.windowFilePath()
        return pathlib.Path(path) if path else None

    def get_image(self) -> gui.QPixmap:
        """Get a pixmap for the widget.

        Implements a workaround to also include QOpenGlWidgets.
        """
        from prettyqt.qt import QtOpenGLWidgets

        image = self.grab()
        with gui.Painter(image) as painter:
            painter.set_composition_mode("source_atop")
            for gl_widget in self.find_children(QtOpenGLWidgets.QOpenGLWidget):
                d = gl_widget.mapToGlobal(core.Point()) - self.mapToGlobal(core.Point())
                painter.drawImage(d, gl_widget.grabFramebuffer())
        return image

    def get_screen(self) -> gui.Screen | None:
        window = self.window().windowHandle()
        return None if window is None else gui.Screen(window.screen())

    def delete_children(self):
        """Delete all children of the specified QObject."""
        if hasattr(self, "clear"):
            return self.clear()
        layout = self.layout()
        while layout.count():
            item = layout.takeAt(0)
            if widget := item.widget():
                widget.deleteLater()

    def get_cursor(self) -> gui.Cursor:
        return gui.Cursor(self.cursor())

    def set_style(self, style: str | widgets.QStyle):
        """Set widget style."""
        if isinstance(style, str):
            style = widgets.QStyleFactory.create(style)
        self.setStyle(style)

    def child_at(self, *args, typ: type[widgets.QWidget] | None = None):
        """Get child widget at position. If type is given, search parents recursively."""
        child = super().childAt(*args)
        if typ is None or isinstance(child, typ):
            return child
        while child := child.parent():
            if isinstance(child, typ):
                return child

    def set_layout_direction(
        self, direction: constants.LayoutDirectionStr | constants.LayoutDirection | None
    ):
        """Set layout direction.

        Args:
            direction: layout direction
        """
        if direction is None:
            self.unsetLayoutDirection()
        else:
            self.setLayoutDirection(constants.LAYOUT_DIRECTION.get_enum_value(direction))

    def get_layout_direction(self) -> constants.LayoutDirectionStr:
        """Get the current layout direction.

        Returns:
            layout direction
        """
        return constants.LAYOUT_DIRECTION.inverse[self.layoutDirection()]


class Widget(WidgetMixin, widgets.QWidget):
    """The base class of all user interface objects."""


if __name__ == "__main__":
    app = widgets.app()
    widget = widgets.TableView()
    meta = widget.get_metaobject()
    prop = meta.get_property("editTriggers")
    print(list(prop.read(widget)))
    widget2 = Widget(status_tip="trekk", whats_this="kfjk", minimum_width=None)
    # widget.play_animation(
    #     "property",
    #     name="windowOpacity",
    #     duration=1000,
    #     start_value=0,
    #     end_value=1,
    # )
    # widget.play_animation("fade_in")
    # val = animations.FadeInAnimation()
    # val.apply_to(widget)
    # val.start()
    widget.fx.set_graphics_effect("colorize")
    widget.show()
    widget2.show()
    mainwindow = widgets.MainWindow()
    container = widgets.Widget()
    container.set_layout("horizontal")
    container.box.add(widget)
    container.box.add(widget2)
    mainwindow.setCentralWidget(container)
    mainwindow.show()
    app.sleep(4)
    # widget.position_on("mouse", scale_ratio=0.5, how="top")
    # widget.set_min_size((400, 400))
    # widget.set_max_size(None, 600)
    app.exec()



from __future__ import annotations

from prettyqt import widgets
from prettyqt.utils import get_repr


class GraphicsTextItem(widgets.GraphicsObjectMixin, widgets.QGraphicsTextItem):
    """Text item that you can add to a QGraphicsScene to display formatted text."""

    def __repr__(self):
        return get_repr(self, self.toPlainText())



from __future__ import annotations

import logging

import mknodes

from prettyqt import core


logger = logging.getLogger(__name__)


class PropertyTable(mknodes.MkTable):
    """A table containing info about Qt properties."""

    def __init__(
        self,
        qobject: type[core.QObject],
        user_prop_name: str | None = None,
        header: str = "",
    ):
        lines = []
        headers = ["Qt Property", "Type", "Doc"]
        properties = core.MetaObject(qobject.staticMetaObject).get_properties()
        doc_dict = core.Property.get_doc_dict(qobject)
        for prop in properties:
            property_name = f"`{prop.get_name()}`"
            if prop.get_name() == user_prop_name:
                property_name += " *(User property)*"
            # if (flag := prop.get_enumerator()):
            meta_type = prop.get_meta_type()
            label = (meta_type.get_name() or "").rstrip("*")
            doc = doc_dict.get(prop.get_name(), "")
            # mark = "x" if prop.get_name() == user_prop_name else ""
            sections = [property_name, f"**{label}**", doc]
            lines.append(sections)
        super().__init__(columns=headers, data=list(zip(*lines)), header=header)


if __name__ == "__main__":
    table = PropertyTable(core.StringListModel)
    dct = core.Property.get_doc_dict(core.StringListModel)
    print(dct)
    # print(table.to_markdown())



from __future__ import annotations

from collections.abc import Sequence
import logging

import pydantic

from prettyqt import constants, core, itemmodels


logger = logging.getLogger(__name__)


class PydanticModel(itemmodels.BaseDataclassModel):
    """Table model to display a list of [pydantic](https://www.pydantic.dev/) BaseModels.

    Non-frozen fields can be edited.

    !!! note
        This ItemModel only supports Pydantic v2 BaseModels.
    """

    SUPPORTS = Sequence[pydantic.BaseModel]

    def __init__(self, items: Sequence[pydantic.BaseModel], **kwargs):
        super().__init__(items, **kwargs)

    @classmethod
    def supports(cls, instance) -> bool:
        match instance:
            case (pydantic.BaseModel(), *_):
                return True
            case _:
                return False

    def get_fields(self) -> list:
        return self.Class.model_fields

    def flags(self, parent: core.ModelIndex) -> constants.ItemFlag:
        if not parent.isValid():
            return super().flags(parent)
        if self.Class.model_config.get("frozen"):
            super().flags(parent)
        field_name = self._field_names[parent.column()]
        instance = self.items[parent.row()]
        # need to cover not parent.isValid()?
        val = getattr(instance, field_name)
        if isinstance(val, bool):
            return super().flags(parent) | constants.IS_CHECKABLE
        else:
            return super().flags(parent) | constants.IS_EDITABLE


if __name__ == "__main__":
    from prettyqt import widgets

    class BaseSetting(pydantic.BaseModel):
        name: str
        label: str
        description: str
        requires_restart: bool = False
        options: list[str] | None = None
        minimum: float = 0.0
        maximum: float = 0.0
        pattern: str = ""

    class SelectionSetting(BaseSetting):
        options: list[str] | None = None
        minimum: float = 0.0
        maximum: float = 0.0
        pattern: str = ""
        requires_restart: bool = False

    app_style = BaseSetting(
        name="app_style",
        label="App style",
        description="Some longer text",
        options=[],
        requires_restart=True,
    )
    app = widgets.app()
    with app.debug_mode():
        items = [
            BaseSetting(name=f"test{i}", label="some label", description="jhaj")
            for i in range(5)
        ]
        view = widgets.TableView()
        view.set_icon("mdi.folder")
        model = PydanticModel(items, parent=view)
        view.set_model(model)
        view.setEditTriggers(view.EditTrigger.AllEditTriggers)
        view.set_delegate("editor")
        view.resize(1000, 1000)
        view.show()
        app.exec()



from __future__ import annotations

from typing import Literal

from prettyqt import constants, core, widgets
from prettyqt.utils import bidict


CompletionModeStr = Literal["popup", "inline", "unfiltered_popup"]

COMPLETION_MODE: bidict[CompletionModeStr, widgets.QCompleter.CompletionMode] = bidict(
    popup=widgets.QCompleter.CompletionMode.PopupCompletion,
    inline=widgets.QCompleter.CompletionMode.InlineCompletion,
    unfiltered_popup=widgets.QCompleter.CompletionMode.UnfilteredPopupCompletion,
)


SortModeStr = Literal["unsorted", "case_sensitive", "case_insensitive"]

SORT_MODE: bidict[SortModeStr, widgets.QCompleter.ModelSorting] = bidict(
    unsorted=widgets.QCompleter.ModelSorting.UnsortedModel,
    case_sensitive=widgets.QCompleter.ModelSorting.CaseSensitivelySortedModel,
    case_insensitive=widgets.QCompleter.ModelSorting.CaseInsensitivelySortedModel,
)


class Completer(core.ObjectMixin, widgets.QCompleter):
    """Completions based on an item model."""

    path_updated = core.Signal(str)

    def _get_map(self):
        maps = super()._get_map()
        maps |= {
            "caseSensitivity": constants.CASE_SENSITIVITY,
            "completionMode": COMPLETION_MODE,
            "modelSorting": SORT_MODE,
            "filterMode": constants.MATCH_FLAGS,
        }
        return maps

    def splitPath(self, path: str) -> list[str]:
        self.path_updated.emit(path)
        return super().splitPath(path)

    def get_completions(self) -> list[str]:
        completions = []
        # count = self.completionCount()
        # for i in range(count):
        #     self.setCurrentRow(i)
        #     completions.append(self.currentCompletion())
        # according to docs, completionCount should be avoided. Not sure if thats true.
        i = 0
        while self.setCurrentRow(i):
            completions.append(self.currentCompletion())
            i += 1
        return completions

    def set_strings(self, strings: list[str]):
        model = core.StringListModel(strings)
        self.setModel(model)

    def set_sort_mode(self, mode: SortModeStr | widgets.QCompleter.ModelSorting | None):
        """Set sort mode to use.

        Args:
            mode: sort mode to use
        """
        if mode is None:
            mode = "unsorted"
        self.setModelSorting(SORT_MODE.get_enum_value(mode))

    def get_sort_mode(self) -> SortModeStr:
        """Return current sort mode.

        Returns:
            sort mode
        """
        return SORT_MODE.inverse[self.modelSorting()]

    def set_completion_mode(
        self, mode: CompletionModeStr | widgets.QCompleter.CompletionMode
    ):
        """Set completion mode to use.

        Args:
            mode: completion mode to use
        """
        self.setCompletionMode(COMPLETION_MODE.get_enum_value(mode))

    def get_completion_mode(self) -> CompletionModeStr:
        """Return current completion mode.

        Returns:
            completion mode
        """
        return COMPLETION_MODE.inverse[self.completionMode()]

    def set_filter_mode(self, mode: constants.FilterModeStr | constants.MatchFlag):
        """Set filter mode to use.

        Args:
            mode: filter mode to use
        """
        self.setFilterMode(constants.FILTER_MODES.get_enum_value(mode))

    def get_filter_mode(self) -> constants.FilterModeStr:
        """Return current filter mode.

        Returns:
            filter mode
        """
        return constants.FILTER_MODES.inverse[self.filterMode()]

    def set_case_sensitive(self, state: bool):
        """Set case sensitivity.

        Args:
            state: case sensitive

        """
        sensitivity = (
            constants.CaseSensitivity.CaseSensitive
            if state
            else constants.CaseSensitivity.CaseInsensitive
        )
        self.setCaseSensitivity(sensitivity)

    def is_case_sensitive(self) -> bool:
        """Return case sensitivity.

        Returns:
            case sensitivity
        """
        return self.caseSensitivity() == constants.CaseSensitivity.CaseSensitive

    def set_completion_role(
        self, role: constants.ItemDataRoleStr | constants.ItemDataRole | int
    ):
        role = constants.ITEM_DATA_ROLE[role] if isinstance(role, str) else role
        self.setCompletionRole(role)


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    completer = Completer()
    model = widgets.FileSystemModel()
    model.set_root_path("")
    app.sleep(10)
    completer.setModel(model)



from __future__ import annotations

import pathlib

from typing_extensions import Self

from prettyqt import core, paths
from prettyqt.utils import datatypes


class TranslatorMixin(core.ObjectMixin):
    def __bool__(self):
        return not self.isEmpty()

    def get_file_path(self) -> pathlib.Path | None:
        path = self.filePath()
        return pathlib.Path(path) if path else None

    def load_file(self, path: datatypes.PathType):
        path = pathlib.Path(path)
        if not self.load(path.name, str(path.parent)):
            raise OSError(f"Invalid language file {path}")

    @classmethod
    def get_available_languages(cls) -> set[str]:
        return {
            str(path).split("_", maxsplit=1)[1][:-3]
            for path in paths.LOCALIZATION_PATH.iterdir()
        }

    @classmethod
    def for_language(cls, language: str) -> Self:
        # if language not in cls.get_available_languages():
        #     raise ValueError("Language does not exist")
        file = paths.LOCALIZATION_PATH / f"language_{language}.qm"
        translator = cls()
        translator.load_file(file)
        return translator

    @classmethod
    def for_system_language(cls) -> Self:
        translator = cls()
        if not translator.load(
            f"qt_{core.Locale.system().name()}",
            str(core.LibraryInfo.get_location("translations")),
        ):
            raise OSError("Could not get translator for system language")
        return translator


class Translator(TranslatorMixin, core.QTranslator):
    """Internationalization support for text output."""


if __name__ == "__main__":
    print(Translator.get_available_languages())
    app = core.app()
    translator = Translator.for_language("de")



from __future__ import annotations

from prettyqt import core, positioning
from prettyqt.qt import QtPositioning
from prettyqt.utils import get_repr


class GeoAreaMonitorInfo(QtPositioning.QGeoAreaMonitorInfo):
    def __str__(self):
        return self.name()

    def __repr__(self):
        return get_repr(self, self.name())

    def get_area(self) -> QtPositioning.QGeoShape:
        area = self.area()
        match area:
            case QtPositioning.QGeoCircle():
                return positioning.GeoCircle(area)
            case QtPositioning.QGeoRectangle():
                return positioning.GeoRectangle(area)
            case QtPositioning.QGeoPath():
                return positioning.GeoPath(area)
            case QtPositioning.QGeoPolygon():
                return positioning.GeoPolygon(area)
            case QtPositioning.QGeoShape():
                return positioning.GeoShape(area)
            case _:
                raise RuntimeError

    def get_expiration(self) -> core.DateTime:
        return core.DateTime(self.expiration())


if __name__ == "__main__":
    info = GeoAreaMonitorInfo("test")
    print(info)
    print(repr(info))



from __future__ import annotations

import logging
import re


_MAXCACHE = 20

logger = logging.getLogger(__name__)


class BaseResolver:
    _match_cache = {}

    def __init__(self, ignore_case: bool = False):
        """Base resolver. Subclass to get glob functionality.

        Keyword Args:
            name (str): Name of the node attribute to be used for resolving
            ignore_case (bool): Enable case insensisitve handling.
        """
        super().__init__()
        self.ignore_case = ignore_case

    def get_parent(self, node):
        return NotImplemented

    def get_children(self, node):
        return NotImplemented

    def get_attribute(self, node):
        return NotImplemented

    def get_root(self, node):
        prev = node
        while node:
            node = self.get_parent(node)
            prev = node
        return prev

    def get_separator(self, node) -> str:
        return "/"

    def get(self, path: str, root_node):
        """Return instance at `path`.

        An example module tree:

        >>> top = Node("top", parent=None)
        >>> sub0 = Node("sub0", parent=top)
        >>> sub0sub0 = Node("sub0sub0", parent=sub0)
        >>> sub0sub1 = Node("sub0sub1", parent=sub0)
        >>> sub1 = Node("sub1", parent=top)

        A resolver using the `name` attribute:

        >>> r = Resolver('name')

        Relative paths:

        >>> r.get(top, "sub0/sub0sub0")
        Node('/top/sub0/sub0sub0')
        >>> r.get(sub1, "..")
        Node('/top')
        >>> r.get(sub1, "../sub0/sub0sub1")
        Node('/top/sub0/sub0sub1')
        >>> r.get(sub1, ".")
        Node('/top/sub1')
        >>> r.get(sub1, "")
        Node('/top/sub1')
        >>> r.get(top, "sub2")
        Traceback (most recent call last):
          ...
        ChildResolverError: Node('/top') has no child sub2.
        Children are: 'sub0', 'sub1'.

        Absolute paths:

        >>> r.get(sub0sub0, "/top")
        Node('/top')
        >>> r.get(sub0sub0, "/top/sub0")
        Node('/top/sub0')
        >>> r.get(sub0sub0, "/")
        Traceback (most recent call last):
          ...
        ResolverError: root node missing. root is '/top'.
        >>> r.get(sub0sub0, "/bar")
        Traceback (most recent call last):
          ...
        ResolverError: unknown root node '/bar'. root is '/top'.

        Going above the root node raises a :any:`RootResolverError`:

        >>> r.get(top, "..")
        Traceback (most recent call last):
            ...
        RootResolverError: Cannot go above root node Node('/top')

        .. note:: Please not that :any:`get()` returned `None` in exactly that case above,
                  which was a bug until version 1.8.1.

        Case insensitive matching:

        >>> r.get(top, '/TOP')
        Traceback (most recent call last):
            ...
        ResolverError: unknown root node '/TOP'. root is '/top'.

        >>> r = Resolver('name', ignore_case=True)
        >>> r.get(top, '/TOp')
        Node('/top')
        """
        node, parts = self.__start(root_node, path, self.__cmp)
        for part in parts:
            if part == "..":
                parent = self.get_parent(node)
                if parent is None:
                    raise RootResolverError(node)
                node = parent
            elif part not in ("", "."):
                node = self.__get(node, part)
        return node

    def __get(self, node, name):
        for child in self.get_children(node):
            if self.__cmp(self.get_attribute(child), str(name)):
                return child
        names = [repr(self.get_attribute(c)) for c in self.get_children(node)]
        raise ChildResolverError(node, name, names)

    def glob(self, path: str, root_node):
        """Return instances at `path` supporting wildcards.

        Behaves identical to :any:`get`, but accepts wildcards and returns
        a list of found nodes.

        * `*` matches any characters, except '/'.
        * `?` matches a single character, except '/'.

        An example module tree:

        >>> top = Node("top", parent=None)
        >>> sub0 = Node("sub0", parent=top)
        >>> sub0sub0 = Node("sub0", parent=sub0)
        >>> sub0sub1 = Node("sub1", parent=sub0)
        >>> sub1 = Node("sub1", parent=top)
        >>> sub1sub0 = Node("sub0", parent=sub1)

        A resolver using the `name` attribute:

        >>> r = Resolver('name')

        Relative paths:

        >>> r.glob(top, "sub0/sub?")
        [Node('/top/sub0/sub0'), Node('/top/sub0/sub1')]
        >>> r.glob(sub1, ".././*")
        [Node('/top/sub0'), Node('/top/sub1')]
        >>> r.glob(top, "*/*")
        [Node('/top/sub0/sub0'), Node('/top/sub0/sub1'), Node('/top/sub1/sub0')]
        >>> r.glob(top, "*/sub0")
        [Node('/top/sub0/sub0'), Node('/top/sub1/sub0')]
        >>> r.glob(top, "sub1/sub1")
        Traceback (most recent call last):
            ...
        ChildResolverError: Node('/top/sub1') has no child sub1.
        Children are: 'sub0'.

        Non-matching wildcards are no error:

        >>> r.glob(top, "bar*")
        []
        >>> r.glob(top, "sub2")
        Traceback (most recent call last):
          ...
        ChildResolverError: Node('/top') has no child sub2.
        Children are: 'sub0', 'sub1'.

        Absolute paths:

        >>> r.glob(sub0sub0, "/top/*")
        [Node('/top/sub0'), Node('/top/sub1')]
        >>> r.glob(sub0sub0, "/")
        Traceback (most recent call last):
          ...
        ResolverError: root node missing. root is '/top'.
        >>> r.glob(sub0sub0, "/bar")
        Traceback (most recent call last):
          ...
        ResolverError: unknown root node '/bar'. root is '/top'.

        Going above the root node raises a :any:`RootResolverError`:

        >>> r.glob(top, "..")
        Traceback (most recent call last):
            ...
        RootResolverError: Cannot go above root node Node('/top')
        """
        node, parts = self.__start(root_node, path, self.__match)
        return self.__glob(node, parts)

    def __start(self, node, path: str, cmp_):
        sep = self.get_separator(node)
        parts = path.split(sep)
        # resolve root
        if path.startswith(sep):
            node = self.get_root(node)
            rootpart = self.get_attribute(node)
            parts.pop(0)
            if not parts[0]:
                msg = f"root node missing. root is '{sep}{rootpart}'."
                raise ResolverError(node, "", msg)
            if not cmp_(rootpart, parts[0]):
                msg = f"unknown root node '{sep}{parts[0]}'. root is '{sep}{rootpart}'."
                raise ResolverError(node, "", msg)
            parts.pop(0)
        return node, parts

    def __glob(self, node, parts):
        assert node is not None
        nodes = []
        if parts:
            name = parts[0]
            remainder = parts[1:]
            # handle relative
            if name == "..":
                parent = self.get_parent(node)
                if parent is None:
                    raise RootResolverError(node)
                else:
                    nodes += self.__glob(parent, remainder)
            elif name in ("", "."):
                nodes += self.__glob(node, remainder)
            elif matches := self.__find(node, name, remainder):
                nodes += matches
            elif self.is_wildcard(name):
                nodes += matches
            else:
                names = [repr(self.get_attribute(c)) for c in self.get_children(node)]
                raise ChildResolverError(node, name, names)
        else:
            nodes = [node]
        return nodes

    def __find(self, node, pat, remainder):
        matches = []
        for child in self.get_children(node):
            name = self.get_attribute(child)
            try:
                if self.__match(name, pat):
                    if remainder:
                        matches += self.__glob(child, remainder)
                    else:
                        matches.append(child)
            except ResolverError as exc:  # noqa: PERF203
                if not self.is_wildcard(pat):
                    raise exc
        return matches

    @staticmethod
    def is_wildcard(path: str) -> bool:
        """Return `True` is a wildcard."""
        return "?" in path or "*" in path

    def __match(self, name, pat):
        k = (pat, self.ignore_case)
        try:
            re_pat = self._match_cache[k]
        except KeyError:
            res = self.__translate(pat)
            if len(self._match_cache) >= _MAXCACHE:
                self._match_cache.clear()
            flags = 0
            if self.ignore_case:
                flags |= re.IGNORECASE
            self._match_cache[k] = re_pat = re.compile(res, flags=flags)
        return re_pat.match(name) is not None

    def __cmp(self, name, pat):
        return name.upper() == pat.upper() if self.ignore_case else name == pat

    @staticmethod
    def __translate(pat):
        re_pat = ""
        for char in pat:
            if char == "*":
                re_pat += ".*"
            elif char == "?":
                re_pat += "."
            else:
                re_pat += re.escape(char)
        return f"(?ms){re_pat}" + r"\Z"  # noqa: ISC003


class ResolverError(RuntimeError):
    def __init__(self, node, child, msg):
        """Resolve Error at `node` handling `child`."""
        super().__init__(msg)
        self.node = node
        self.child = child


class RootResolverError(ResolverError):
    def __init__(self, root):
        """Root Resolve Error, cannot go above root node."""
        msg = f"Cannot go above root node {root!r}"
        super().__init__(root, None, msg)


class ChildResolverError(ResolverError):
    def __init__(self, node, child, names):
        """Child Resolve Error at `node` handling `child`."""
        msg = "{!r} has no child {}. Children are: {}.".format(
            node, child, ", ".join(names)
        )
        super().__init__(node, child, msg)


if __name__ == "__main__":
    from prettyqt import core, itemmodels, widgets

    app = widgets.app()
    model = itemmodels.ParentClassTreeModel(core.AbstractAnimation)
    widget = widgets.TreeView()
    widget.set_model(model)
    resolver = itemmodels.ItemModelResolver(model)
    print("here")
    # widget.show()
    # app.exec()
    test = resolver.glob("/")
    print(test)



from __future__ import annotations

from prettyqt import constants, widgets
from prettyqt.qt import QtMultimediaWidgets


# QtMultimediaWidgets.QVideoWidget.__bases__ = (
#     widgets.Widget,
#     multimedia.MediaBindableInterface,
# )


class VideoWidget(widgets.WidgetMixin, QtMultimediaWidgets.QVideoWidget):
    def __init__(self, *args, **kwargs):
        self.doubleclick_for_fullscreen = False
        super().__init__(*args, **kwargs)

    def set_doubleclick_for_fullscreen(self, value: bool = True):
        self.doubleclick_for_fullscreen = value

    def mouseDoubleClickEvent(self, event):
        if (
            event.button() == constants.MouseButton.LeftButton
            and self.doubleclick_for_fullscreen
        ):
            self.setFullScreen(not self.isFullScreen())
            event.accept()
        return super().mouseDoubleClickEvent(event)


if __name__ == "__main__":
    app = widgets.app()
    widget = VideoWidget()
    widget.show()
    app.exec()



from __future__ import annotations

from collections.abc import Callable
import functools
import logging
import operator

from typing import TYPE_CHECKING, Any, Literal

from prettyqt import constants, core, gui, widgets
from prettyqt.utils import classhelpers, datatypes


if TYPE_CHECKING:
    from prettyqt import itemmodels


logger = logging.getLogger(__name__)


ProxyStr = Literal[
    "fuzzy",
    "transpose",
    "sort_filter",
    "identity",
    "value_transformation",
    "range_filter",
    "checkable",
    "subset",
    "flatten_tree",
    "table_to_list",
    "predicate_filter",
    "size_limiter",
    "subsequence",
    "appearance",
    "column_join",
    "read_only",
    "highlight_current",
]


class Sliced:
    def __init__(self, indexer, widget: widgets.QAbstractItemView, proxifier: Proxifier):
        if widget.model() is None:
            raise RuntimeError("Need a model in order to proxify.")
        # PySide6 shows empty tables when no parent is set.
        if widget.model().parent() is None:
            widget.model().setParent(widget)
        self._indexer = indexer
        self.proxifier = proxifier
        self._widget = widget

    def filter(self) -> itemmodels.SliceFilterProxyModel:
        """Filter subsection to display.

        Wraps current model with a SliceFilterProxyModel.
        """
        from prettyqt import itemmodels

        proxy = itemmodels.SliceFilterProxyModel(
            indexer=self._indexer,
            parent=self._widget,
            source_model=self._widget.model(),
        )
        self._widget.set_model(proxy)
        return proxy

    def highlight_current(
        self, mode="column"
    ) -> itemmodels.SliceHighlightCurrentProxyModel:
        """Filter subsection to display.

        Wraps current model with a SliceHighlightCurrentProxyModel.
        """
        from prettyqt import itemmodels

        proxy = itemmodels.SliceHighlightCurrentProxyModel(
            indexer=self._indexer,
            parent=self._widget,
            mode=mode,
            source_model=self._widget.model(),
        )
        self._widget.set_model(proxy)
        self._widget.selectionModel().currentChanged.connect(proxy.highlight_index)
        return proxy

    def modify(
        self,
        fn: Callable[[Any], Any],
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
        selector: Callable[[Any], bool] | None = None,
        selector_role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ) -> itemmodels.SliceValueTransformationProxyModel:
        """Conditionally apply modifications to given area.

        Wraps current model with a SliceValueTransformationProxyModel.

        Arguments:
            fn: Callable to use for modifying.
            role: role to modify
            selector: Callable to filter cells which should be modified.
            selector_role: data role the selector callable should get as an argument.
        """
        from prettyqt import itemmodels

        proxy = itemmodels.SliceValueTransformationProxyModel(
            indexer=self._indexer,
            parent=self._widget,
            source_model=self._widget.model(),
        )
        proxy.add_transformer(fn, role, selector, selector_role)
        self._widget.set_model(proxy)
        return proxy

    def change_flags(
        self,
        selectable: bool | None = None,
        editable: bool | None = None,
        drag_enabled: bool | None = None,
        drop_enabled: bool | None = None,
        user_checkable: bool | None = None,
        enabled: bool | None = None,
        auto_tristate: bool | None = None,
        never_has_children: bool | None = None,
        user_tristate: bool | None = None,
    ) -> itemmodels.SliceChangeFlagsProxyModel:
        """Change Item flags for given slice.

        For makin an area checkable, usually set_checkable should be preferred
        since it keeps track of checked items and triggers a callback on checkstate
        change.

        Arguments:
            selectable: Changes the ItemIsSelectable Flag
            editable: Changes the ItemIsEnabled Flag
            drag_enabled: Changes the ItemIsDragEnabed Flag
            drop_enabled: Changes the ItemIsDropEnabled Flag
            user_checkable: Changes the ItemIsUserCheckable Flag
            enabled: Changes the ItemIsEnabled Flag
            auto_tristate: Changes the ItemIsAutoTristate Flag
            never_has_children: Changes the ItemNeverHasChildren Flag
            user_tristate: Changes the ItemIsUserTristate Flag
        """
        from prettyqt import itemmodels

        proxy = itemmodels.SliceChangeFlagsProxyModel(
            indexer=self._indexer,
            parent=self._widget,
            source_model=self._widget.model(),
        )
        flags = dict(
            selectable=selectable,
            editable=editable,
            drag_enabled=drag_enabled,
            drop_enabled=drop_enabled,
            user_checkable=user_checkable,
            enabled=enabled,
            auto_tristate=auto_tristate,
            never_has_children=never_has_children,
            user_tristate=user_tristate,
        )
        flags_to_add = [constants.ITEM_FLAG[k] for k, v in flags.items() if v is True]
        flags_to_remove = [constants.ITEM_FLAG[k] for k, v in flags.items() if v is False]
        if flags_to_add:
            proxy.set_flags_to_add(functools.reduce(operator.ior, flags_to_add))
        if flags_to_remove:
            proxy.set_flags_to_remove(functools.reduce(operator.ior, flags_to_remove))
        self._widget.set_model(proxy)
        return proxy

    def set_format(
        self,
        int_format: str | None = None,
        float_format: str | None = None,
        datetime_format: str | None = None,
        date_format: str | None = None,
        time_format: str | None = None,
    ) -> itemmodels.SliceDisplayTextProxyModel:
        """Format non-str values returned by DisplayRole.

        Wraps current model with a SliceDisplayTextProxyModel.

        Arguments:
            int_format: Format to use for int values
            float_format: Format to use for float values
            datetime_format: Format to use for QDateTime / datetime.datetime objects.
            date_format: Format to use for QDate / datetime.date objects
            time_format: Format to use for QTime / datetime.time objects.
        """
        from prettyqt import itemmodels

        proxy = itemmodels.SliceDisplayTextProxyModel(
            int_format=int_format or "{:.4f}",
            float_format=float_format or "{:.4f}",
            datetime_format=datetime_format or "%m/%d/%Y, %H:%M:%S",
            date_format=date_format or "%m/%d/%Y",
            time_format=time_format or "%H:%M:%S",
            source_model=self._widget.model(),
        )
        self._widget.set_model(proxy)
        return proxy

    def set_checkable(
        self,
        callback: Callable[[core.ModelIndex], Any] | None = None,
        tree: bool = False,
    ) -> itemmodels.SliceCheckableProxyModel | itemmodels.SliceCheckableTreeProxyModel:
        """Make given area checkable and trigger a callback on checkstate change.

        For trees, set tree=True. That way checkboxes change to tristate and
        propagate changes to parents and children.

        Arguments:
            callback: Callback to trigger when checkstate changes.
            tree: Whether the underlying model is a tree.
        """
        from prettyqt import itemmodels

        if tree:
            proxy = itemmodels.SliceCheckableTreeProxyModel(
                indexer=self._indexer,
                parent=self._widget,
                source_model=self._widget.model(),
            )
        else:
            proxy = itemmodels.SliceCheckableProxyModel(
                indexer=self._indexer,
                parent=self._widget,
                source_model=self._widget.model(),
            )
        if callback:
            proxy.checkstate_changed.connect(callback)
        self._widget.set_model(proxy)
        return proxy

    def change_icon_size(
        self, size: datatypes.SizeType
    ) -> itemmodels.SliceChangeIconSizeProxyModel:
        """Change the size of pixmap / icon provided by decoration role.

        Wraps current model with a  SliceChangeIconSizeProxyModel.

        Arguments:
            size: New size for decoration role.
        """
        from prettyqt import itemmodels

        size = datatypes.to_size(size)
        proxy = itemmodels.SliceChangeIconSizeProxyModel(
            indexer=self._indexer,
            size=size,
            parent=self._widget,
            source_model=self._widget.model(),
        )
        self._widget.set_model(proxy)
        return proxy

    def style(
        self,
        foreground: datatypes.ColorType | gui.QBrush | None = None,
        background: datatypes.ColorType | gui.QBrush | None = None,
        font: str | gui.QFont | None = None,
        alignment: constants.AlignmentFlag | constants.AlignmentStr | None = None,
    ) -> itemmodels.SliceAppearanceProxyModel:
        """Apply styling to given area.

        Wraps current model with a SliceAppearanceProxyModel.

            foreground: Color / Brush to use for foreground role.
            background: Color / Brush to use for background role.
            font: Font to use for font role.
            alignment: Alignment to use for alignment role.
        """
        from prettyqt import itemmodels

        proxy = itemmodels.SliceAppearanceProxyModel(
            indexer=self._indexer,
            parent=self._widget,
            source_model=self._widget.model(),
        )
        proxy.set_foreground(foreground)
        proxy.set_background(background)
        proxy.set_font(font)
        proxy.set_alignment(alignment)
        self._widget.set_model(proxy)
        return proxy

    def map_role(
        self,
        from_: constants.ItemDataRole,
        to: constants.ItemDataRole,
        converter: Callable | None = None,
    ) -> itemmodels.SliceMapRoleProxyModel:
        """Map ItemDataRole to another role for data().

        Wraps model with a MapRoleProxyMpodel.

        Arguments:
            from_: role to map from
            to: role to map to
            converter: modify mapped values with callable
        """
        from prettyqt import itemmodels

        proxy = itemmodels.SliceMapRoleProxyModel(
            indexer=self._indexer,
            parent=self._widget,
            mapping={from_: to},
            converter=converter,
            source_model=self._widget.model(),
        )
        self._widget.set_model(proxy)
        return proxy

    def color_values(
        self,
        low_color: datatypes.ColorType = "green",
        high_color: datatypes.ColorType = "red",
    ) -> itemmodels.SliceColorValuesProxyModel:
        """Color numerical values.

        Wraps model with a SliceColorValuesProxyModel.

        Arguments:
            low_color: color to use for "low" values
            high_color: color to use for "high" values
        """
        from prettyqt import itemmodels

        proxy = itemmodels.SliceColorValuesProxyModel(
            indexer=self._indexer, parent=self._widget, source_model=self._widget.model()
        )
        proxy.set_low_color(low_color)
        proxy.set_high_color(high_color)
        self._widget.set_model(proxy)
        return proxy

    def color_categories(
        self,
    ) -> itemmodels.SliceColorCategoriesProxyModel:
        """Color numerical values.

        Wraps model with a SliceColorCategoriesProxyModel.
        """
        from prettyqt import itemmodels

        proxy = itemmodels.SliceColorCategoriesProxyModel(
            indexer=self._indexer, parent=self._widget, source_model=self._widget.model()
        )
        self._widget.set_model(proxy)
        return proxy


class Proxifier:
    """The proxifier object offers a more declarative way to set up proxy models.

    ``` py
    model = MyTableModel()
    table = widgets.TableView()
    table.set_model(model)

    # lets change the appearance a bit.
    # Set the font color of column 2 and 3 to red and font to Courier.

    table.proxifier[:, 2:4].style(foreground="red", font="Courier")

    # Cut off last column and only show last 50 lines.
    table.proxifier[:50, :-1].filter()

    # Set first 20 lines of these 50 lines to read_only
    table.proxifier[:20, :].change_flags(editable=False)

    # Make first column checkable and trigger callback on checkstate change.
    table.proxifier[0].make_checkable(callback=my_callback)
    ```
    """

    def __init__(self, widget: widgets.QAbstractItemView):
        self._widget = widget
        self._wrapper = None

    def __getitem__(self, value: slice) -> Sliced:
        """Return a Sliced Object."""
        logger.debug(f"Building {value!r} ProxyModel for {self._widget!r}")
        self._wrapper = Sliced(indexer=value, widget=self._widget, proxifier=self)
        return self._wrapper

    def __getattr__(self, name: str):
        indexer = (slice(None, None, None), slice(None, None, None))
        self._wrapper = Sliced(indexer=indexer, widget=self._widget, proxifier=self)
        if hasattr(self._wrapper, name):
            return getattr(self._wrapper, name)
        else:
            raise AttributeError(name)

    def transpose(self) -> core.TransposeProxyModel:
        """Transpose rows/columns.

        Wraps current model with a TransposeProxyModel.
        """
        proxy = core.TransposeProxyModel(
            parent=self._widget, source_model=self._widget.model()
        )
        self._widget.set_model(proxy)
        return proxy

    def flatten(
        self, show_path: bool = False, leaves_only: bool = False
    ) -> itemmodels.FlattenTreeProxyModel:
        """Wraps model in a Proxy which flattens tree to one column.

        Arguments:
            show_path: Whether the first column should show the full tree path.
            leaves_only: whether the proxied model should return only tree leaves.

        """
        # ss = """QTreeView::branch{border-image: url(none.png);}"""
        # self._widget.set_stylesheet(ss)
        from prettyqt import itemmodels

        proxy = itemmodels.FlattenTreeProxyModel(
            parent=self._widget,
            show_path=show_path,
            leaves_only=leaves_only,
            source_model=self._widget.model(),
        )
        self._widget.set_model(proxy)
        return proxy

    def melt(
        self, id_columns: list[int], var_name: str = "Variable", value_name: str = "Value"
    ) -> itemmodels.MeltProxyModel:
        """Wraps model in a Proxy which unpivots the table to a long format.

        Arguments:
            id_columns: Identifier variables
            var_name: header to use for variable column
            value_name: header to use for value_name
        """
        from prettyqt import itemmodels

        proxy = itemmodels.MeltProxyModel(
            id_columns=id_columns,
            var_name=var_name,
            value_name=value_name,
            parent=self._widget,
            source_model=self._widget.model(),
        )
        self._widget.set_model(proxy)
        return proxy

    def reorder_columns(self, order: list[int]) -> itemmodels.ColumnOrderProxyModel:
        """Reorder columns to given order.

        Wraps current model with a ColumnOrderProxyModel which rearranges columns to given
        order.

        Arguments:
            order: list of indexes. Does not need to include all column indexes,
                  missing ones will be hidden.
        """
        from prettyqt import itemmodels

        proxy = itemmodels.ColumnOrderProxyModel(
            order=order, parent=self._widget, source_model=self._widget.model()
        )
        self._widget.set_model(proxy)
        return proxy

    def to_list(self) -> itemmodels.TableToListProxyModel:
        """Convert table to a list.

        Wraps model with a TableToListProxyModel which reshapes table to one column
        by concatenating the seperatate columns into a one-columned list.
        """
        from prettyqt import itemmodels

        proxy = itemmodels.TableToListProxyModel(
            parent=self._widget, source_model=self._widget.model()
        )
        self._widget.set_model(proxy)
        return proxy

    def add_column(
        self,
        header: str,
        formatter: str,
        flags: constants.ItemFlag | None = None,
    ) -> itemmodels.ColumnJoinerProxyModel:
        """Add a new column with given header to the table.

        Column content can be defined by a formatter.

        Arguments:
            header: Title for section header
            formatter: String formatter (Example: "{2} - {4}" would result in
                       <displayRole of column 2> - <displayRole of column4>
            flags: ItemFlags for new column (default: Enabled and selectable)
        """
        from prettyqt import itemmodels

        proxy = itemmodels.ColumnJoinerProxyModel(
            parent=self._widget, source_model=self._widget.model()
        )
        proxy.add_mapping(header=header, formatter=formatter, flags=flags)
        self._widget.set_model(proxy)
        return proxy

    def change_headers(
        self,
        headers: list[Any] | dict[int, Any],
        orientation: constants.Orientation
        | constants.OrientationStr = constants.HORIZONTAL,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ) -> itemmodels.ChangeHeadersProxyModel:
        """Change headers of source model.

        Wraps current model with a ChangeHeadersProxyModel.

        Arguments:
            headers: new headers to use
            orientation: orientation of the header which should be modified.
            role: Header role to change
        """
        from prettyqt import itemmodels

        proxy = itemmodels.ChangeHeadersProxyModel(
            header=headers,
            role=role,
            orientation=orientation,
            parent=self._widget,
            source_model=self._widget.model(),
        )
        self._widget.set_model(proxy)
        return proxy

    def set_sort_filter_proxy(self, **kwargs) -> core.SortFilterProxyModel:
        proxy = core.SortFilterProxyModel(
            parent=self._widget, source_model=self._widget.model(), **kwargs
        )
        self._widget.set_model(proxy)
        return proxy

    def get_proxy(self, proxy: ProxyStr, **kwargs) -> core.QAbstractProxyModel:
        Klass = classhelpers.get_class_for_id(core.AbstractProxyModelMixin, proxy)
        proxy_instance = Klass(
            parent=self._widget, source_model=self._widget.model(), **kwargs
        )
        self._widget.set_model(proxy_instance)
        return proxy_instance

    def map_to(
        self,
        index_or_selection: core.ModelIndex | core.QItemSelection,
        target: widgets.QAbstractItemView | core.QAbstractItemModel,
    ) -> core.ModelIndex | core.QItemSelection:
        """Map index or selection to given target.

        Arguments:
            index_or_selection: What should be mapped.
            target: Either an ItemView or a (proxy) model which is linked to our current
                    model.
        """
        if isinstance(target, widgets.QAbstractItemView):
            target = target.model()
        mapper = itemmodels.ProxyMapper(self._widget.model(), target)
        match index_or_selection:
            case core.ModelIndex():
                return mapper.map_index(source=0, target=1, index=index_or_selection)
            case core.QItemSelection():
                return mapper.map_selection(
                    source=0, target=1, selection=index_or_selection
                )
            case _:
                raise TypeError(index_or_selection)

    def map_from(
        self,
        index_or_selection: core.ModelIndex | core.QItemSelection,
        source: widgets.QAbstractItemView | core.QAbstractItemModel,
    ) -> core.ModelIndex | core.QItemSelection:
        """Map index or selection from given source.

        Arguments:
            index_or_selection: What should be mapped.
            source: Either an ItemView or a (proxy) model which is linked to our current
                    model.
        """
        if isinstance(source, widgets.QAbstractItemView):
            source = source.model()
        mapper = itemmodels.ProxyMapper(self._widget.model(), source)
        match index_or_selection:
            case core.ModelIndex():
                return mapper.map_index(source=1, target=0, index=index_or_selection)
            case core.QItemSelection():
                return mapper.map_selection(
                    source=1, target=0, selection=index_or_selection
                )
            case _:
                raise TypeError(index_or_selection)

    def sync_current_selection_with(
        self,
        target: widgets.QAbstractItemView | core.QAbstractItemModel,
    ):
        """Map index or selection from given source.

        Arguments:
            target: Either an ItemView or a (proxy) model which is linked to our current
                    model.
        """
        if isinstance(target, widgets.QAbstractItemView):
            target = target.model()
        itemmodels.ProxyMapper(self._widget.model(), target)


if __name__ == "__main__":
    from prettyqt import debugging

    app = widgets.app()

    table = debugging.example_table()
    table.proxifier[0].change_flags(selectable=False, enabled=False)
    table.show()
    with app.debug_mode():
        app.exec()



from __future__ import annotations

from typing import Literal

from prettyqt import constants, core, itemmodels
from prettyqt.qt import QtGui
from prettyqt.utils import colors, datatypes


HighlightModeStr = Literal["column", "row", "all"]


class SliceHighlightCurrentProxyModel(itemmodels.SliceIdentityProxyModel):
    """Proxy model which highlights all cells with same data as current index.

    Highlights all cells with same content in given role as currently focused cell
    with a user-specified color.

    Possible modes are:

    * `all`: Highlight all cells with same value.
    * `column`: Highlight all cells with same value and same column as current.
    * `row`: Highlight all cells with same value and same row as current.

    === "Without proxy"

        ```py
        dct = dict(
            a=["a", "b", "a", "b"],
            b=["a", "b", "a", "b"],
            c=["a", "b", "a", "b"],
            d=["b", "a", "b", "a"],
            e=["a", "b", "a", "a"],
        )
        model = gui.StandardItemModel.from_dict(dct)
        table = widgets.TableView()
        table.set_model(model)
        # apply proxy to every 2nd column
        # table.proxifier[:, ::2].highlight_current(mode="column")
        ```
        <figure markdown>
          ![Image title](../../images/slicehighlightcurrentproxymodel_none.png)
        </figure>

    === "Row mode"

        ```py
        dct = dict(
            a=["a", "b", "a", "b"],
            b=["a", "b", "a", "b"],
            c=["a", "b", "a", "b"],
            d=["b", "a", "b", "a"],
            e=["a", "b", "a", "a"],
        )
        model = gui.StandardItemModel.from_dict(dct)
        table = widgets.TableView()
        table.set_model(model)
        # apply proxy to every 2nd column
        table.proxifier[:, ::2].highlight_current(mode="row")
        ```
        <figure markdown>
          ![Image title](../../images/slicehighlightcurrentproxymodel_row.png)
        </figure>

    === "Column mode"

        ```py
        dct = dict(
            a=["a", "b", "a", "b"],
            b=["a", "b", "a", "b"],
            c=["a", "b", "a", "b"],
            d=["b", "a", "b", "a"],
            e=["a", "b", "a", "a"],
        )
        model = gui.StandardItemModel.from_dict(dct)
        table = widgets.TableView()
        table.set_model(model)
        # apply proxy to every 2nd column
        table.proxifier[:, ::2].highlight_current(mode="column")
        ```
        <figure markdown>
          ![Image title](../../images/slicehighlightcurrentproxymodel_column.png)
        </figure>

    === "All mode"

        ```py
        dct = dict(
            a=["a", "b", "a", "b"],
            b=["a", "b", "a", "b"],
            c=["a", "b", "a", "b"],
            d=["b", "a", "b", "a"],
            e=["a", "b", "a", "a"],
        )
        model = gui.StandardItemModel.from_dict(dct)
        table = widgets.TableView()
        table.set_model(model)
        # apply proxy to every 2nd column
        table.proxifier[:, ::2].highlight_current(mode="all")
        ```
        <figure markdown>
          ![Image title](../../images/slicehighlightcurrentproxymodel_all.png)
        </figure>


    ### Example

    ```py
    model = MyModel()
    table = widgets.TableView()
    table.set_model(model)
    table[:, :3].proxify.highlight_current(mode="all")
    table.show()
    # or
    indexer = (slice(None), slice(None, 3))
    proxy = itemmodels.SliceFilterProxyModel(indexer=indexer)
    proxy.set_source_model(model)
    table.set_model(proxy)
    table.selectionModel().currentChanged.connect(proxy.highlight_index)
    table.show()
    ```
    """

    ID = "highlight_current"

    def __init__(
        self,
        role=constants.DISPLAY_ROLE,
        mode: HighlightModeStr = "column",
        highlight_color: datatypes.ColorType = "red",
        **kwargs,
    ):
        self._mode = mode
        self._current_value = ...  # Sentinel value
        self._data_role = role
        self._current_column = None
        self._current_row = None
        self._highlight_color = colors.get_color(highlight_color).as_qt()
        super().__init__(**kwargs)

    def set_highlight_color(self, color: datatypes.ColorType):
        """Set color used for highlighting cells."""
        self._highlight_color = colors.get_color(color).as_qt()

    def get_highlight_color(self) -> QtGui.QColor:
        """Get color used for higlighting cells."""
        return self._highlight_color

    def set_highlight_mode(self, mode: HighlightModeStr):
        """Set highlight mode."""
        self._highlight_mode = mode

    def get_highlight_mode(self) -> HighlightModeStr:
        """Get highlight mode."""
        return self._highlight_mode

    def set_highlight_role(self, mode: constants.ItemDataRole):
        """Set highlight mode."""
        self._data_role = mode

    def get_highlight_role(self) -> constants.ItemDataRole:
        """Get highlight mode."""
        return self._data_role

    def set_highlight_column(self, column: int):
        with self.change_layout():
            self._current_column = column

    def get_highlight_column(self) -> int:
        return self._current_column

    def set_highlight_row(self, row: int):
        with self.change_layout():
            self._current_row = row

    def get_highlight_row(self) -> int:
        return self._current_row

    def set_current_value(self, value):
        with self.change_layout():
            self._current_value = value

    def highlight_index(self, index: core.ModelIndex):
        with self.change_layout():
            self._current_value = index.data(self._data_role)  # super().data(index, role)
            self._current_row = index.row()
            self._current_column = index.column()

    def data(
        self,
        index: core.ModelIndex,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ):
        if (
            role == constants.BACKGROUND_ROLE
            and index.data(self._data_role) == self._current_value
            and self.indexer_contains(index)
            and (
                (self._mode == "column" and index.column() == self._current_column)
                or (
                    self._mode == "row"
                    and (index.row() == self._current_row)
                    or self._mode == "all"
                )
            )
        ):
            return self._highlight_color
        return super().data(index, role)

    highlightMode = core.Property(
        str,
        get_highlight_mode,
        set_highlight_mode,
        doc="Highlight mode",
    )
    highlightColor = core.Property(
        QtGui.QColor,
        get_highlight_color,
        set_highlight_color,
        doc="Color to use for highlighting",
    )
    highlightRole = core.Property(
        constants.ItemDataRole,
        get_highlight_role,
        set_highlight_role,
        doc="ItemRole to use for highlighting",
    )
    highlight_column = core.Property(
        int,
        get_highlight_column,
        set_highlight_column,
        doc="Currently highlighted column",
    )
    highlight_row = core.Property(
        int,
        get_highlight_row,
        set_highlight_row,
        doc="Currently highlighted row",
    )


if __name__ == "__main__":
    from prettyqt import gui, widgets

    app = widgets.app(style="Windows")
    dct = dict(
        a=["a", "b", "a", "b"],
        b=["a", "b", "a", "b"],
        c=["a", "b", "a", "b"],
        d=["b", "a", "b", "a"],
        e=["a", "b", "a", "a"],
    )
    model = gui.StandardItemModel.from_dict(dct)
    table = widgets.TableView()
    table.set_model(model)
    table.proxifier[:, ::2].highlight_current(mode="column")
    table.resize(500, 200)
    table.set_title("Example")
    table.set_icon("mdi.cursor-default-click-outline")
    table.show()
    table.h_header.resize_sections("stretch")
    with app.debug_mode():
        app.exec()
